//  Copyright (c) 1997-2000 by Mark F. Adams 
//  $Id: gauss_seid.C,v 1.31 2006/08/10 17:20:56 adams Exp $
//  Author:             Mark F. Adams
//  Version:            2.0
//  Filename:           gauss_seid.C
//

#include "prometheus.hh"
#include <limits.h>

#define HASHP(p) (p%fact1*fact2 + p/fact1)
//#define HASHP(p) (p)

// local helper function - get (block) lid with local equation
#define GETLBIDWITHGEQ(geqi) ((geqi<my0eq || geqi>=next0eq) ? -1 : (geqi - my0eq)/bs_)

#undef __FUNCT__  
#define __FUNCT__ "PromPCGaussSeidel::PromPCGaussSeidel"
/* PromPCGaussSeidel::PromPCGaussSeidel() ********************************
 * 
 */ 
PromPCGaussSeidel::PromPCGaussSeidel( const PromOptions &opt, 
				      PromPerfMonitor &perf ) :
  PromPC_base(opt,perf), 
  maxnGSBlocks_(0),
  nGSBlocks_(0),
  bs_(0),
  pre_smooth_(TRUE),
  symmetric_(FALSE),
  comm_(MPI_COMM_NULL),
  NodalFactors_(NULL),
  ghostx_(NULL), 
  NodalFactorPointers_(NULL),
  pc_(NULL), 
  A_(NULL),
  gbid_ghost_(9769),
  lowProc_send_(10),
  lowProc_recv_(10), 
  highProc_recv_(10),
  highProc_send_(10),
  gbid_ISBlock_(253),
  proc_color_(53)
{
  for( int ii = 0 ; ii < PROM_NBLKLISTTYPES ; ii++ ){
    assert(blocksLists_[ii].array_ == NULL);
    assert(blocksLists_[ii].arr_size_ == 0);
  }
}

#undef __FUNCT__  
#define __FUNCT__ "PromPCGaussSeidel::Create"
/* PromPCGaussSeidel::Create() ********************************
 * 
 *    - forms 'gbid_ghost_' -- the only methods that interact with 
 *                             multi-blocks and ERF mapping
 */ 
int PromPCGaussSeidel::Create( const PromMatrix * const AA, 
			       const PromIS * const domainISs,
			       const int nblocks ) 
{
  int ierr,tt[2],nb[2],jj;
  ierr = MatGetBlockSize( AA->mat_, &bs_ ); CHKERRQ( ierr ); 
  ierr = MatGetOwnershipRange( AA->mat_, &tt[0], &tt[1] ); CHKERRQ( ierr );
  const int nneq=tt[1]-tt[0];
  comm_ =  AA->map_.comm_(); const int fact = AA->isComplex() ? 2 : 1;
  A_ = AA;           assert(bs_ > 0);
  
  // constructor -- use matrix blocks
  nGSBlocks_ = (domainISs==NULL) ? nneq/bs_ : nblocks;
  nb[0] = nGSBlocks_; nb[1] = (pc_ != NULL);
  MPI_Allreduce( &nb, &tt, 2, MPI_INT, MPI_MAX, comm_ );
  maxnGSBlocks_ = tt[0]; assert(nb[1] == (pc_ != NULL));
  
  perf_mon_.EventBegin(PromContext::prom_logs_[LOG_MIS2]); 
  if( comm_ != MPI_COMM_NULL ) { 
    int *garr, ng;
    ierr = AA->GetGArray( &garr, &ng );      CHKERRQ(ierr);
    assert( gbid_ghost_.isEmpty() );
    for( jj = 0; jj < ng ; /*void*/ ) {
      int gbid = garr[jj];
      if( AA->map_.isConstNDF() ) {
	// add
	PromGSNode *node = new PromGSNode( jj, gbid );
	int add = gbid_ghost_.add( node, gbid+1 ); assert(add==1);
	jj++;
      }
      else{
	const int gid = AA->ghostBID_GID_->find(gbid+1) - 1;  assert(gid>=0);
	const int ndf = AA->ghostGID_ndf_->find(gid+1)*fact;  assert(ndf>0);
	const int nn = ndf/bs_; assert(ndf%bs_==0);
	// add a g-s 'block' for each sub block
	while( jj<ng && garr[jj]>=gbid && garr[jj]<gbid+nn ){  
	  PromGSNode *node = new PromGSNode( jj, gid, nn );
	  int add = gbid_ghost_.add( node, garr[jj]+1 );  assert(add==1);
	  jj++;
	}
      }
    }
    assert(jj==ng);
    ierr = AA->RestoreGArray(&garr);      CHKERRQ(ierr);
    // setup
    ierr = Create_private( domainISs );    CHKERRQ(ierr); 
  }
  else nGSBlocks_ = 0;
  perf_mon_.EventEnd(PromContext::prom_logs_[LOG_MIS2]);

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromPCGaussSeidel::~PromPCGaussSeidel"
/* PromPCGaussSeidel::~PromPCGaussSeidel() ********************************
 * 
 */ 
PromPCGaussSeidel::~PromPCGaussSeidel()
{
  PromProcXComm *commx;

  for( int ii = 0 ; ii < PROM_NBLKLISTTYPES ; ii++ ){ 
    if( blocksLists_[ii].arr_size_ > 0 ){
      int *arr = blocksLists_[ii].array_; 
      PetscFree(arr);
    }
  }
  while( !lowProc_send_.isEmpty() ) delete lowProc_send_.removeHead(); 
  while( !highProc_recv_.isEmpty() ){ 
    commx = highProc_recv_.removeHead(); 
#if !defined(PARCH_ascired) // not implemented on red
    MPI_Cancel( &commx->req_ ); // *1*
#endif
    delete commx; 
  }
  while( !highProc_send_.isEmpty() ) delete highProc_send_.removeHead(); 
  while( !lowProc_recv_.isEmpty() ){ 
    commx = lowProc_recv_.removeHead(); 
#if !defined(PARCH_ascired) // not implemented on red
    MPI_Cancel( &commx->req_ ); // *1*
#endif
    delete commx; 
  }
  while( !gbid_ISBlock_.isEmpty() ) delete gbid_ISBlock_.removeHead(); 
  if( NodalFactors_ != NULL ) PetscFree( NodalFactors_ );
  if( ghostx_ != NULL ) assert(ghostx_[gbid_ghost_.getCount()*bs_] == -12.);
  if( ghostx_ != NULL ) PetscFree( ghostx_ );
  //if( lid_geq_ != NULL ) PetscFree( lid_geq_ );
  while( !gbid_ghost_.isEmpty() ) delete gbid_ghost_.removeHead();
  //if( ghostLid_locEQ_ != NULL ) PetscFree( ghostLid_locEQ_ );
  if( NodalFactorPointers_ != NULL ) PetscFree( NodalFactorPointers_ );
  if(pc_) delete pc_;
}

#undef __FUNCT__  
#define __FUNCT__ "PromPCGaussSeidel::SetUp_private"
/* PromPCGaussSeidel::SetUp_private **************************************
 *
 */     
int PromPCGaussSeidel::SetUp_private()
{
  int ierr; 
  PetscFunctionBegin;
#if defined(PROM_TRACE_PC_SETUP)
  PetscPrintf(comm_,"\t[%d]%s this=%p\n",A_->map_.comm_.mype(),__FUNCT__,this);
#endif
  
  if( pc_ != NULL ) {
    ierr = pc_->SetUp(); CHKERRQ(ierr);
  }
  else {
    ierr = SetupNodalGS(); CHKERRQ(ierr); 
  }
  if( ghostx_ != NULL ){
    int xx = gbid_ghost_.getCount() * bs_; 
    while(xx--)ghostx_[xx] = PROM_BIG_NUM;
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__  
#define __FUNCT__ "PromPCGaussSeidel::Create_private"
/* PromPCGaussSeidel::Create_private ********************************
 *
 *   - Gauss-Seidel smoother
 *
 *   INPUT:
 *    - A:
 *    - domainISs:
 *
 *   SIDE EFFECTS:
 *
 *   RETURN:
 *     - PETSc error code
 */
int PromPCGaussSeidel::Create_private( const PromIS * const domainISs )
{
  int ierr,tt,ii,*lid_blockGID;
  ierr = MatGetOwnershipRange( A_->mat_, &ii, &tt); CHKERRQ( ierr );
  const int myb0=ii/bs_,nextb0=tt/bs_;
  const int np=A_->map_.comm_.np(),mype=A_->map_.comm_.mype();
  PromTable<PromTable<int>*> proc_boundLids(57);
  PromBlockType    *block_type;
  PromTable<int>  **block_gid_ghostLid;
  PromList<int>   **block_procs, **block_nodeLids; 
  PromList<int>   **block_boundLids, **block_adjacBLids;

  PetscFunctionBegin;
  // DoBlocks data: myb0(node)
  for( ii= 0 ; ii < PROM_NBLKLISTTYPES ; ii++ ){ 
    assert(blocksLists_[ii].list_ == NULL);
    assert(blocksLists_[ii].arr_size_ == 0);
    blocksLists_[ii].list_ = new PromList<int>(12);
    CHKNEW(blocksLists_[ii].list_,1);
  }
  tt = nGSBlocks_+1; ii = sizeof(PromList<int>*);
  ierr = PetscMalloc(tt*ii,&block_procs);
  CHKERRQ(ierr);  
  PetscMemzero(block_procs,tt*ii);
  ierr = PetscMalloc(tt*ii,&block_gid_ghostLid); CHKERRQ(ierr);  
  PetscMemzero(block_gid_ghostLid,tt*ii);
  ierr = PetscMalloc(tt*ii,&block_nodeLids);     CHKERRQ(ierr);
  PetscMemzero(block_nodeLids,tt*ii);
  ierr = PetscMalloc(tt*ii,&block_boundLids);    CHKERRQ(ierr);  
  PetscMemzero(block_boundLids,tt*ii);
  ierr = PetscMalloc(tt*ii,&block_adjacBLids);   CHKERRQ(ierr);  
  PetscMemzero(block_adjacBLids,tt*ii);
  ierr = PetscMalloc(tt*sizeof(PromBlockType),& block_type); CHKERRQ(ierr); 
  for( ii = 0 ; ii < nGSBlocks_ ; ii++ ) block_type[ii] = PROM_INTER;

  { // setup data structures
    PromList<int> highlist, lowlist; 
    ierr = Setup_private(block_procs, block_gid_ghostLid,block_nodeLids,
    			 block_boundLids, block_adjacBLids, &proc_boundLids,
			 domainISs, &highlist, &lowlist );
    CHKERRQ(ierr); tt = gbid_ghost_.getCount() * bs_;
    ierr = PetscMalloc((tt+1)*sizeof(double), &ghostx_);  CHKERRQ(ierr);
    ghostx_[tt] = -12.;
    // color processors
    ierr = ColorProcs( &highlist, &lowlist ); CHKERRQ(ierr); 
  }
  ii = (nextb0-myb0) + gbid_ghost_.getCount();
  ierr = PetscMalloc((ii+1)*sizeof(int),&lid_blockGID);    CHKERRQ(ierr);
  while( ii-- ) lid_blockGID[ii] = -1;

  // set types -- add to 'gbid_ISBlock'
  ierr = SetBlockTypes( block_procs, &proc_boundLids, block_type,
			lid_blockGID, domainISs );
  CHKERRQ(ierr); 

  // make buffers
  ierr = MakeLists( block_procs, block_gid_ghostLid, block_type, 
		    lid_blockGID, block_nodeLids, block_boundLids,
		    block_adjacBLids, domainISs );
  CHKERRQ(ierr); 

  // clean-up
  for( ii = 0 ; ii< nGSBlocks_ ; ii++ ) {
    if( block_procs[ii] ) { 
      delete block_procs[ii]; 
      delete block_gid_ghostLid[ii]; 
      delete block_boundLids[ii]; 
    }
    else assert(block_boundLids[ii]==NULL);
    delete block_nodeLids[ii]; 
    delete block_adjacBLids[ii]; 
  }
  PetscFree(block_procs);       PetscFree(block_gid_ghostLid); 
  PetscFree(block_type);        PetscFree(lid_blockGID); 
  PetscFree(block_nodeLids);    PetscFree(block_boundLids);
  PetscFree(block_adjacBLids);
  while ( !proc_boundLids.isEmpty() ) {
    PromTable<int> *tab = proc_boundLids.removeHead();
    delete tab; 
  }

  // output 
  static int vt_count = 0;  // first grid only so no blocking!!!
  if( options_.verbose_ > 2 && vt_count++ == 0 ) { 
    char typenames[PROM_NBLKLISTTYPES][21] = { 
      "                Infs",    "              INTER1", 
      "             INTER1B",    " INF_SUP_WO_HIGHERIS", 
      "            INF_SUPL",    "  INF_SUP_WO_LOWERIS", 
      "              INTER2",    "             INTER2B",
      "                Sups" };
    int lcounts[PROM_NBLKLISTTYPES], gcounts[PROM_NBLKLISTTYPES];
    for(ii=0;ii<PROM_NBLKLISTTYPES;ii++){
      lcounts[ii] = blocksLists_[ii].list_->getCount();
    }
    Position pos = lowProc_recv_.getHeadPosition(); 
    while( pos ){
      PromProcXComm *commx = lowProc_recv_.getNext(pos);
      lcounts[PROM_INFL] += commx->InfBlocks_.list_->getCount();
    }
    pos = highProc_recv_.getHeadPosition();  
    while( pos ){
      PromProcXComm *commx = highProc_recv_.getNext(pos);
      lcounts[PROM_SUPL] += commx->InfBlocks_.list_->getCount();
    }
    MPI_Allreduce(lcounts,gcounts,PROM_NBLKLISTTYPES,MPI_INT,MPI_SUM,comm_);
    PetscPrintf(comm_,"\t\tGLOBAL Block counts (nLocNds_=%d):\n",
		A_->map_.nlocal());
    for( int x = 0 ; x < PROM_NBLKLISTTYPES ; x++ ){
      PetscPrintf( comm_, " %s : %5d. ", typenames[x], gcounts[x] );
      if( (x+1)%4 == 0 ) PetscPrintf(comm_,"\n");
    }
    PetscPrintf(comm_,"\n");
    for( int p = 0 ; p < np ; p++ ){
      MPI_Barrier( comm_ );
      //PetscSleep(1);
      if( p == mype ) {
	PetscPrintf(PETSC_COMM_SELF,"[%d]Block List: %d blocks, %d procs\n",
		    mype,nGSBlocks_,
		    lowProc_send_.getCount()+highProc_send_.getCount());
	for( int x = 0 ; x < PROM_NBLKLISTTYPES ; x++ ){
	  PetscPrintf(PETSC_COMM_SELF, " %s : %5d. ",typenames[x], lcounts[x]);
	  if( (x+1)%4 == 0 ) PetscPrintf(PETSC_COMM_SELF,"\n");
	}
	PetscPrintf(PETSC_COMM_SELF,"\n");
      }
    }
  }

  // conver to arrays
  PromList<PromProcXComm*> *lists[4] = { &lowProc_send_, &lowProc_recv_,
					 &highProc_recv_, &highProc_send_ };
  for( int kk = 0 ; kk < 4 ; kk++ ) {
    PromList<PromProcXComm*> *list = lists[kk];
    Position pos = list->getHeadPosition(); 
    while( pos ){
      PromProcXComm *commx = list->getNext( pos );
      int *arr = NULL, sz = commx->InfBlocks_.list_->getCount(), jj = 0; 
      if( sz > 0 ) {
	ierr = PetscMalloc(sizeof(int)*(sz+1),&arr);  CHKERRQ(ierr);
	Position pos2 = commx->InfBlocks_.list_->getHeadPosition(); 
	while( pos2 ) arr[jj++] = commx->InfBlocks_.list_->getNext(pos2) - 1;
	assert(jj==sz); 
      }
      delete commx->InfBlocks_.list_;	
      commx->InfBlocks_.list_ = NULL;
      commx->InfBlocks_.array_ = arr;
      commx->InfBlocks_.arr_size_ = sz;
    }
  }
  for( ii = 0 ; ii<PROM_NBLKLISTTYPES ; ii++){
    assert(blocksLists_[ii].arr_size_ == 0);
    if( blocksLists_[ii].list_ != NULL ) {
      int *arr = NULL, sz = blocksLists_[ii].list_->getCount(), jj = 0; 
      if( sz > 0 ) {
	ierr = PetscMalloc(sizeof(int)*(sz+1),&arr);  CHKERRQ(ierr);
	Position pos = blocksLists_[ii].list_->getHeadPosition(); 
	while( pos ) arr[jj++] = blocksLists_[ii].list_->getNext( pos ) - 1;
	assert(jj==sz); 
      }
      delete blocksLists_[ii].list_;	
      blocksLists_[ii].list_ = NULL;
      blocksLists_[ii].array_ = arr;
      blocksLists_[ii].arr_size_ = sz;
    }
  }
  
  PetscFunctionReturn(0);
}

#undef __FUNCT__  
#define __FUNCT__ "PromPCGaussSeidel::ReorderNodes"
/* PromPCGaussSeidel::ReorderNodes ***********************************
 *
 *   order nodes 
 *
 *   INPUT:
 *    - wieghts[len]:
 *    - len
 *
 *   SIDE EFFECTS:
 *
 *   RETURN:
 *     - PETSc error code
 */ 
int PromPCGaussSeidel::ReorderNodes( double wieghts[], int len )
{
  const int ndsz = ( sizeof(float) + sizeof(int) ) / sizeof(int);
  int ierr, kk, ii, maxIS_id, buffsz = PROM_BSZ, *buff, *pp;
  const int mype = A_->map_.comm_.mype();
  int __buff[PROM_BSZ * ndsz];

  if( pc_ != NULL ) SETERRQ(1,"ReorderNodes not implemented for blocks");
  buff = __buff;
  for( kk = maxIS_id = 0 ; kk<PROM_NBLKLISTTYPES ; kk++ ){
    PromGSCont *list = &blocksLists_[kk];
    int *arr = list->array_, sz = list->arr_size_;
    // buffer
    if( sz > buffsz ) {
      if( buff != __buff ) PetscFree(buff);
      buffsz = 2 * sz;
      ierr = PetscMalloc( buffsz*ndsz*sizeof(int), &buff );  CHKERRQ(ierr); 
    }
    // construct array to sort
    for( ii = 0, pp = buff ; ii < sz ; ii++, pp += ndsz ){
      const int lidi = arr[ii]; 
      float *fp = (float*)pp; *fp++ = wieghts[lidi]; 
      int *ppp = (int*)fp;	*ppp = lidi;
    }
    // sort
    qsort( buff, sz, ndsz*sizeof(int), prom_floatcompar );
    // repack
    pp = buff + sizeof(float)/sizeof(int);
    for( ii = 0; ii < sz ; ii++, pp += ndsz ) arr[ii] = *pp;
    // get 'maxIS_id'
    if( kk == PROM_INF_SUP_WO_HIGHERIS || kk == PROM_INF_SUPL || 
	kk == PROM_INF_SUP_WO_LOWERIS ) {
      for(ii=0;ii<sz;ii++)if(arr[ii]>maxIS_id) maxIS_id = arr[ii];
    }
  }
  if( maxIS_id > buffsz ) {
    if( buff != __buff ) PetscFree(buff);
    buffsz = maxIS_id;
    ierr = PetscMalloc( buffsz*ndsz*sizeof(int), &buff );  CHKERRQ(ierr); 
  }

  // 
  for( ii = 0; ii < buffsz ; ii++, pp += ndsz ) buff[ii] = -1;  
  // set order in 'buff' 
  for( kk = 0 ; kk < PROM_NBLKLISTTYPES ; kk++ ){
    if( kk == PROM_INF_SUP_WO_HIGHERIS || 
	kk == PROM_INF_SUPL || 
	kk == PROM_INF_SUP_WO_LOWERIS ) {
      PromGSCont *list = &blocksLists_[kk];
      int *arr = list->array_, sz = list->arr_size_, jump;
      switch(kk){
      case PROM_INF_SUP_WO_HIGHERIS : jump = 2*maxnGSBlocks_; break;
      case PROM_INF_SUPL :            jump = maxnGSBlocks_; break;
      case PROM_INF_SUP_WO_LOWERIS :  jump = 0; break;
      default: assert(0); 
      }
      for( ii = 0 ; ii < sz ; ii++ ) buff[arr[ii]] = ii + jump;
    }
  }

  // redo 'high_ISBlocks_' and 'low_ISBlocks_'
  for( kk = 0 ; kk<PROM_NBLKLISTTYPES ; kk++ ){
    if( kk == PROM_INF_SUP_WO_HIGHERIS || 
	kk == PROM_INF_SUPL || 
	kk == PROM_INF_SUP_WO_LOWERIS ) {
      PromGSCont *list = &blocksLists_[kk];
      int *arr = list->array_, sz = list->arr_size_;
      for( ii = 0 ; ii < sz ; ii++ ){
	const int blid = arr[ii], iind = buff[blid]; assert(iind!=-1);
	int gbid = mype*maxnGSBlocks_ + blid;
	PromGSBlock *isb = gbid_ISBlock_.find(gbid+1);
	PromInfSupBlock *ISlocal = dynamic_cast<PromInfSupBlock*>(isb);
	assert(ISlocal!=NULL);
	// 'high_ISBlocks_'
	Position pos2,pos = ISlocal->high_ISBlocks_.getHeadPosition(); 
	while( (pos2=pos) ){ 
	  PromGSBlock *b2 = ISlocal->high_ISBlocks_.getNext(pos);
	  if( b2->getProc() == mype ) {
	    const int gid = b2->getGBID(), jblid = gid%maxnGSBlocks_;
	    const int jind = buff[jblid]; 
	    assert(gid >= mype*maxnGSBlocks_ && gid < (mype+1)*maxnGSBlocks_);
	    assert(jind != iind); assert(jind != -1);
	    if( jind < iind ) { // "<"
	      ISlocal->high_ISBlocks_.removeAt(pos2);
	      ISlocal->low_ISBlocks_.addTail(b2);
	    }
	  }
	}
	// 'low_ISBlocks_'
	pos = ISlocal->low_ISBlocks_.getHeadPosition(); 
	while( (pos2=pos) ){ 
	  PromGSBlock *b2 = ISlocal->low_ISBlocks_.getNext(pos);
	  if( b2->getProc() == mype ) {
	    const int gid=b2->getGBID(),jblid=gid%maxnGSBlocks_,jind=buff[jblid]; 
	    assert(gid >= mype*maxnGSBlocks_ && gid < (mype+1)*maxnGSBlocks_);
	    assert(jind != iind);assert(jind != -1);
	    if( jind > iind ) { // ">"
	      ISlocal->high_ISBlocks_.removeAt(pos2);
	      ISlocal->low_ISBlocks_.addTail(b2);
	    }
	  }
	}
      }
    }
  }

  if( buff != __buff ) PetscFree(buff);

  return 0;
}


#undef __FUNCT__  
#define __FUNCT__ "PromPCGaussSeidel::Setup_private"
/* PromPCGaussSeidel::Setup_private ***********************************
 *
 *   Set up data structures
 *
 *   INPUT:
 *    - block_procs[nblocks]: processors that each block talks to (out)
 *    - block_gid_ghostLid[nblocks]: ghosts that each block talks to (out)
 *    - block_nodeLids[nblocks]: lid of nodes in block (out);
 *    - block_boundLids[nblocks]: lid of boundary nodes in block (out)
 *    - block_adjacBLids[nblocks]: local block ids of adjacs of block (out)
 *    - proc_boundLids: proc -->table of bound nodes that touch 'proc' (out) 
 *    - domainISs: 
 *    - highlist: (out)
 *    - lowlist: (out)
 *
 *   SIDE EFFECTS:
 *
 *   RETURN:
 *     - PETSc error code
 */ 
int PromPCGaussSeidel::Setup_private( PromList<int> ** const block_procs,
				      PromTable<int>**const block_gid_ghostLid,
				      PromList<int> **block_nodeLids,
				      PromList<int> **block_boundLids,
				      PromList<int> **block_adjacBLids,
				      PromTable<PromTable<int>*> 
				      * const proc_boundLids,
				      const PromIS * const domainISs,
				      PromList<int> *highlist,
				      PromList<int> *lowlist )
{
  int       ierr,ii,fact1,fact2,sz,jj; const PromMap &map = A_->map_;
  const int *ids;
  int       ncols,blocki,hashp,proc,tid;
  ierr = MatGetOwnershipRange( A_->mat_, &ii, &jj ); CHKERRQ( ierr );
  const int my0eq=ii,myb0=ii/bs_,next0eq=jj,nextb0=jj/bs_;
  const int mype=map.comm_.mype(),np=map.comm_.np();
  const int nbloc=nextb0-myb0, nbnodes = gbid_ghost_.getCount() + nbloc;
  PromTable<int>  blockGids;  
  PromTable<int> *tab;
  int adjacs[PROM_MAXNIEGH]; 
  int *lid_Lbid = NULL, num_nn_max = 1, buff__[PROM_BSZ];

  PetscFunctionBegin;
  // buffer
  if( nbnodes > PROM_BSZ ) {
    ierr = PetscMalloc(nbnodes*sizeof(int),&lid_Lbid); CHKERRQ(ierr);  
  }
  else lid_Lbid = buff__;

  // setup hash function -- get factors -- fact1, fact2
  for( ii = (int)(sqrt( (double)np )) ; ii > 0 ; ii-- ) {
    if( np%ii == 0 ){ fact1 = ii; break; }
  } 
  assert(ii > 0); 
  fact1 *= 1; 
  fact2 = np/fact1;

  // set temp table of node blocks
  const int myhashedPID=HASHP(mype);
  if( domainISs != NULL ) {
    // make and set 'block_nodeLids', set 'num_nn_max'
    for( ii = 0 ; ii < nbnodes ; ii++ ) lid_Lbid[ii] = 0;
    for( blocki = 0, jj = 0 ; blocki < nGSBlocks_ ; blocki++ ) {
      PromList<int> *list = block_nodeLids[blocki] = new PromList<int>(9); 
      CHKNEW( list, 1 );
      const PromIS block = domainISs[blocki];
      ierr = PromISGetIndices( block, &ids );               CHKERRQ(ierr);
      ierr = PromISGetSize( block, &sz ); CHKERRQ(ierr);
      for( ii = 0 ; ii < sz ; ) {
	const int leq = ids[ii], lbid=GETLBIDWITHGEQ(leq); assert(leq%bs_==0);
	if( lbid == -1 ) SETERRQ1(1,"equation %d not found",leq);
	lid_Lbid[lbid]++; assert(lbid<nbnodes);
 	if( lid_Lbid[lbid] > jj ) jj = lid_Lbid[lbid];
	list->addTail(lbid+1); // set block_nodeLids[] -- out
	ii += bs_;
      }
      ierr = PromISRestoreIndices( block, &ids ); CHKERRQ(ierr);
    }
    // make 'lid_Lbid'
    num_nn_max = jj;
    if( (jj=num_nn_max*nbnodes) > PROM_BSZ ) {
      if( lid_Lbid != buff__ ) PetscFree( lid_Lbid );  
      ierr = PetscMalloc(jj*sizeof(int),&lid_Lbid); CHKERRQ(ierr);  
    }
    else lid_Lbid = buff__;
    for( ii = 0 ; ii < jj ; ii++ ) lid_Lbid[ii] = -1;
    // set 'lid_Lbid'
    for( blocki = 0 ; blocki < nGSBlocks_ ; blocki++ ) {
      const PromIS block = domainISs[blocki];
      ierr = PromISGetIndices( block, &ids );               CHKERRQ(ierr);
      ierr = PromISGetSize( block, &sz ); CHKERRQ(ierr);
      for( ii = 0 ; ii < sz ; ) {
	const int leq = ids[ii], lbid = GETLBIDWITHGEQ(leq); 
	if( lbid == -1 ) SETERRQ1(1,"equation %d not found",leq);
	if( lbid >= nbnodes ) SETERRQ1(1,"lbid >= nbnodes %d",lbid);
	int kk, jj = lbid * num_nn_max;
	for( kk = 0 ; kk < num_nn_max ; kk++, jj++ ) {
	  if( lid_Lbid[jj] == -1 ) { lid_Lbid[jj] = blocki; break; }
	  else if( lid_Lbid[jj] == blocki ) SETERRQ(1,"blocki here");
	}
	if(kk==num_nn_max)SETERRQ(1,"");
	ii += bs_;
      }
      ierr = PromISRestoreIndices( block, &ids );   CHKERRQ(ierr);
    }
  }
  else { // lidi == blocki
    num_nn_max = 1;
    for(blocki = 0 ; blocki < nGSBlocks_ ; blocki++) {
      block_nodeLids[blocki] = new PromList<int>(1); 
      CHKNEW(block_nodeLids[blocki],1);
      lid_Lbid[blocki] = blocki;                 // set lid_Lbid
      block_nodeLids[blocki]->addTail(blocki+1); // set block_nodeLids[] -- out
    }
  }

  // out: block_procs,block_gid_ghostLid,proc_boundLids,block_nodeLids,
  //      block_boundLids,block_adjacBLids
  // cloring data: highlist and lowlist (out)
  assert(proc_boundLids->isEmpty()); 
  for( blocki = 0 ; blocki < nGSBlocks_; blocki++ ) {
    PromIS block = NULL;
    // get 'sz', 'ids' and fill 'blockGids' used for iteration below
    if( domainISs != NULL ) {
      block = domainISs[blocki];
      ierr = PromISGetIndices( block, &ids );               CHKERRQ(ierr);
      ierr = PromISGetSize( block, &sz );               CHKERRQ(ierr);
      // form block table
      assert( blockGids.isEmpty() );
      for( ii = 0 ; ii < sz ;	ii += bs_ ) {
	const int lid = GETLBIDWITHGEQ( ids[ii] );
	if( lid == -1 ) SETERRQ1(1,"equation %d not found", ids[ii] );
	// temp table of block nodes
	blockGids.add( lid + myb0 + 1 ); 
      }
    }
    else { // lidi = blocki
      tid = blocki*bs_ + my0eq; ids = &tid; // 1st global equation
      int gbid = blocki + myb0;
      blockGids.add( gbid + 1 ); // temp table of block nodes
      sz = bs_;
    }
    // doit
    for( ii = 0 ; ii < sz ; ii += bs_ ) {
      const int leq = ids[ii], lbidi = GETLBIDWITHGEQ(leq); 
      if(lbidi == -1) { SETERRQ1(1,"gid %d not found",leq); }
      // for all adjacs of node
      if( lbidi < nbloc ) { // local
	ierr = A_->GetLocal_constBlk_NodeRow(lbidi,PROM_MAXNIEGH,ncols,adjacs);
	CHKERRQ(ierr);
      }
      else { // need to add ghosts adajacs!!!
	ncols = 1; adjacs[0] = lbidi + myb0; // not supported
	SETERRQ(1,"");
      }
      for( jj = 0 ; jj < ncols ; jj++ ) {
	const int gbidj = adjacs[jj];
	if( blockGids.findi(gbidj+1) ) continue;    // skip nodes in this block
	if( gbidj >= myb0 && gbidj < nextb0 ) {     // local neigh
	  if( block_adjacBLids[blocki] == NULL ){   // this (blocki) has a ghost
	    block_adjacBLids[blocki] = new PromList<int>(ncols);
	  }
	  const int lbidj = gbidj - myb0;
	  for( int kk = 0, id = num_nn_max*lbidj;
	       kk < num_nn_max && lid_Lbid[id] != -1 ; 
	       kk++, id++ ) {
	    int blockj = lid_Lbid[id];
	    if( !block_adjacBLids[blocki]->find(blockj+1) ) {
	      block_adjacBLids[blocki]->addTail(blockj+1); // out
	    }
	  }
	}
	else {   // a ghost touches this block
	  assert(np>1);
	  if( block_procs[blocki] == NULL ) {  // this (blocki) is a ghost
	    block_procs[blocki] = new PromList<int>(6);
	    CHKNEW(block_procs[blocki],1);
	    block_boundLids[blocki] = new PromList<int>(4); 
	    CHKNEW(block_boundLids[blocki],1);
	    block_gid_ghostLid[blocki] = new PromTable<int>(73);
	    CHKNEW(block_gid_ghostLid[blocki],1);
	  }
	  if( !block_boundLids[blocki]->find(lbidi+1) ) {
	    block_boundLids[blocki]->addTail(lbidi+1); // out
	  }
	  PromGSNode *nodej = gbid_ghost_.find(gbidj+1); 
	  assert(nodej != NULL);
	  const int lbidj = nodej->b.lid_;
	  block_gid_ghostLid[blocki]->add( lbidj+1, gbidj+1 ); // out
	  ierr = GetProc( gbidj, proc ); CHKERRQ(ierr);
	  if( !block_procs[blocki]->find(proc+1) ){ 
	    block_procs[blocki]->addTail(proc+1); // out
	  }
	  if( (tab=proc_boundLids->find(proc+1)) == NULL ) {
	    int tt = gbid_ghost_.getCount() / ((np>10) ? (np>100) ? 8 : 4 : 1);
	    tab = new PromTable<int>(tt + 17); CHKNEW(tab,1);
	    proc_boundLids->add( tab, proc+1 ); // out
	    // set local high/low lists
	    hashp = HASHP(proc); 
	    if( hashp > myhashedPID ) highlist->addTail(proc+1);
	    else lowlist->addTail(proc+1);
	  }
	  tab->add( lbidi + 1 ); // list of boudary nodes for other proc (out)
	}
      }
    }
    blockGids.removeAll();
    if( domainISs != NULL ) {
      ierr = PromISRestoreIndices( block, &ids );           CHKERRQ(ierr);
    }
  }
  if( lid_Lbid != buff__ ) PetscFree(lid_Lbid);  
    
  PetscFunctionReturn(0); 
}

#undef __FUNCT__  
#define __FUNCT__ "PromPCGaussSeidel::ColorProcs"
/* PromPCGaussSeidel::ColorProcs ***********************************
 *
 *   color procs
 *
 *   INPUT:
 *    - highlist: higher procs to comm with 
 *    - lowlist: lower procs to comm with
 *
 *   SIDE EFFECTS:
 *    - forms 'proc_color_' PromTable<int>
 *
 *   RETURN:
 *     - PETSc error code
 */ 
int PromPCGaussSeidel::ColorProcs( PromList<int> * const highlist, 
				   PromList<int> * const lowlist )
{  
  MPI_Status    status; assert(A_!=NULL);
  int           color,isdeleted,mycolor,proc,buff[2];
  const int mype = A_->map_.comm_.mype();
  PromList<int> highlist2; const int tag1 = A_->map_.comm_.getNewTag(comm_);
  Position      pos, pos2; const int tag2 = A_->map_.comm_.getNewTag(comm_);

  // color -- forms 'proc_color_' (side effect)
  PetscFunctionBegin;  
  for( color=1,mycolor=0 ; !highlist->isEmpty() || mycolor==0 ; color++ ) {
    isdeleted = FALSE;
    pos = highlist->getHeadPosition(); 
    while( (pos2=pos) ) {
      proc = highlist->getNext( pos ) - 1; 
      // recv(tag1) from up < PROM_SELECTED | PROM_DELETED, [color] > 
      MPI_Recv( buff, 2, MPI_INT, proc, tag1, comm_, &status);
      int state = buff[0], clr = buff[1];
      if( state == (int)PROM_SELECTED ) {
	// no more recvs from here -- save color -- remove upper
	highlist->removeAt( pos2 );
	highlist2.addTail( proc+1 );
	proc_color_.add( clr, proc+1 );
	if( isdeleted == FALSE && mycolor == 0 ){
	  isdeleted = TRUE;
	  buff[0] = PROM_DELETED; buff[1] = -1;
	  // send(tag1) down < PROM_DELETED, -1 >  
	  Position pos3 = lowlist->getHeadPosition(); 
	  while( pos3 ) {
	    proc = lowlist->getNext( pos3 ) - 1;
	    MPI_Send( buff, 2, MPI_INT, proc, tag1, comm_ );
	  }
	}
      }
    }
    if( isdeleted == FALSE && mycolor == 0 ) { 
      mycolor = color;
      //selected -- send(tag1) down < PROM_SELECTED, color > 
      proc_color_.add( mycolor, mype+1 );
      buff[0] = PROM_SELECTED; buff[1] = mycolor; 
      pos = lowlist->getHeadPosition(); 
      while( pos ) {
	proc = lowlist->getNext( pos ) - 1;
	MPI_Send( buff, 2, MPI_INT, proc, tag1, comm_ );
      }
    }
  }
  assert(mycolor>0);
  if( options_.verbose_ > 2 ){
    PetscPrintf( options_.verbose_ > 1 ? PETSC_COMM_SELF : comm_,
		 "\t\t[%d]%s |Mid|=%d mycolor=%d ncolor=%d %s\n",
		 mype,__FUNCT__,gbid_ISBlock_.getCount(),
		 mycolor, color-1, 
		 (pc_ != NULL) ? "Block G-S" : "nodal G-S" );
  }

  // send(tag2) up < color >
  buff[0] = mycolor;
  pos = highlist2.getHeadPosition(); 
  while( pos ) {
    proc = highlist2.getNext( pos ) - 1;
    MPI_Send( buff, 1, MPI_INT, proc, tag2, comm_ );
  }

  // recv(tag2) from lower  < color >
  while( !lowlist->isEmpty() ) {
    proc = lowlist->removeHead() - 1;  // counter only
    // recv(tag2) from up < color > 
    MPI_Recv( buff, 1, MPI_INT, MPI_ANY_SOURCE, tag2, comm_, &status);
    proc = status.MPI_SOURCE;
    color = buff[0];
    proc_color_.add( color, proc+1 );
  }
  PetscFunctionReturn(0); 
}

#undef __FUNCT__  
#define __FUNCT__ "PromPCGaussSeidel::SetBlockTypes"
/* PromPCGaussSeidel::SetBlockTypes ***********************************
 *
 *   INPUT:
 *    - block_procs[nblocks]: processors that each block talks to 
 *    - proc_boundLids: proc --> table of bound nodes that touch 'proc'
 *    - block_type[nblocks]: (out)
 *    - lid_blockGID[nloc+nghosts]: (out)
 *    - domainISs: 
 *
 *   SIDE EFFECTS:
 *    - sets send/recv buffers: 'lowProc_send_' and 'highProc_recv_'
 *    - adds local PromInfSupBlocks to 'gbid_ISBlock'
 *
 *   RETURN:
 *     - PETSc error code
 */
int PromPCGaussSeidel::SetBlockTypes(const PromList<int> *const*const
				     block_procs,
				     const PromTable<PromTable<int>*>*
				     const proc_boundLids,
				     PromBlockType block_type[],
				     int lid_blockGID[],
				     const PromIS * const domainISs )
{
  int ierr,proc,sz,color,ii,tt,tg,jj,bidi,neq; 
  const int *ids;
  const PromComm_base &comm = A_->map_.comm_; 
  const int np = comm.np(), mype = comm.mype(), mycolor = proc_color_.find(mype+1);
  ierr = MatGetOwnershipRange( A_->mat_, &ii, &jj); CHKERRQ( ierr );
  const int my0eq=ii,myb0=my0eq/bs_,next0eq=jj,nbloc=(next0eq-my0eq)/bs_;
  TablePos tpos, tpos2;  const int tag = comm.getNewTag(comm_); 
  PromBlockType btype;   const int hightag = comm.getNewTag(comm_); 
  MPI_Status status;     const int lowtag = comm.getNewTag(comm_); 
  const int reqsz = sizeof(MPI_Request)/sizeof(int);// +faces
  PromList<MPI_Request*> smessages; 
  const PromList<int> *ilist;
  PromTable<int> *tab;
  
  // catagorize, set local 'lid_blockGID' and 'gbid_ISBlock'
  PetscFunctionBegin;  
  for( bidi = 0 ; bidi < nGSBlocks_; bidi++ ) {
    // set type
    if( (ilist=block_procs[bidi]) ) { // shared 
      jj = np; sz = -1;
      Position pos = ilist->getHeadPosition();  assert(pos!=NULL);
      while( pos ) {
	proc = ilist->getNext( pos ) - 1; 
	color = proc_color_.find(proc+1); assert(color && color!=mycolor);
	if( color < jj ) jj = color; // jj == smallest
	if( color > sz ) sz = color; // sz == largest
      }
      if ( sz > mycolor ) { // have a larger
	if( jj < mycolor ) btype = PROM_INF_SUP; // have a smaller -- mixed
	else btype = PROM_SUP; // no smaller -- I'm on bottom (all superior) 
      }
      else btype = PROM_INF; // no larger (all inferior neighbors)
    }
    else btype = PROM_INTER;
    block_type[bidi] = btype;
    // set IS nodes
    if( btype == PROM_INF_SUP ) {
      int bGID = mype * maxnGSBlocks_ + bidi;
      if( domainISs != NULL ) {
	const PromIS bis = domainISs[bidi]; 
	ierr = PromISGetIndices( bis, &ids );  CHKERRQ(ierr);
	ierr = PromISGetSize( bis, &sz );      CHKERRQ(ierr);
	for( ii = 0 ; ii < sz ;  ) {
	  int lid = GETLBIDWITHGEQ( ids[ii] );
	  if( lid == -1 ) SETERRQ1(1,"equation %d not found", ids[ii] );
	  lid_blockGID[lid] = bGID; // local IS 
	  ii += bs_;
	}
	ierr = PromISRestoreIndices( bis, &ids );           CHKERRQ(ierr);
      }
      else { // lid == bidi
	lid_blockGID[bidi] = bGID; // local IS 
      }
      // create IS_BLOCK and add to list
      PromInfSupBlock *ISlc = new PromInfSupBlock(bGID,mype);CHKNEW(ISlc,1);
      int add = gbid_ISBlock_.add(ISlc, bGID+1);              assert(add>0);
    }
  }

  // isend local (boundary) nodes, add to low(high)Proc_send
  tpos = proc_boundLids->getHeadPosition();
  while( tpos ) {
    tab = proc_boundLids->getNext( tpos, &proc ); proc--; 
    color = proc_color_.find(proc+1);                          assert(color>0);
    tt = 2*tab->getCount();
    MPI_Request *sreq; 
    ierr = PetscMalloc((tt+reqsz)*sizeof(int),&sreq); CHKERRQ(ierr);
    smessages.addTail( sreq );
    int *buff = (int*)(sreq) + reqsz, *pb = buff;
    tpos2 = tab->getHeadPosition();  assert(tpos2!=0); neq = 0;
    while( tpos2 ) {
      int lbid = tab->getNext( tpos2 ) - 1; assert(lbid>=0 && lbid<nbloc);
      *pb++ = lbid + myb0; // gid
      neq += bs_;
      *pb++ = lid_blockGID[lbid];
    }
    assert(tt==pb-buff); 
    // send(tag) *< gbid, -1 | ISgid >
    qsort( buff, tab->getCount(), 2*sizeof(int), prom_intcompar );
    MPI_Isend( buff, tt, MPI_INT, proc, tag, comm_, sreq );
    // make proc send list
    sz = tab->getCount(); assert(sz>0); 
    tg = (color < mycolor) ? lowtag : hightag;
    PromProcXComm *commx = new PromProcXComm(sz,neq,proc,comm_,tg); 
    CHKNEW(commx,1);
    for( ii = 0, tt = 0 ; ii < sz ; ii++, tt += 2 ){
      commx->lids_[ii] = buff[tt] - myb0; // lid for sends
    }
    if( color < mycolor ) lowProc_send_.addTail( commx );
    else highProc_send_.addTail( commx );
  } 

  // recv and make processor x communicators and 'high[/low]Proc_recv_'
  int todo = gbid_ghost_.getCount();  
  while( todo ) {
    assert(todo >= 0);
    MPI_Probe( MPI_ANY_SOURCE, tag, comm_, &status );
    MPI_Get_count( &status, MPI_INT, &tt ); assert(tt > 0);
    int *buff; ierr = PetscMalloc(tt*sizeof(int),&buff); int *pb = buff; 
    CHKERRQ(ierr);
    // recv(tag) *< gbid, -1 | ISgid >
    MPI_Recv( buff, tt, MPI_INT, status.MPI_SOURCE, tag, comm_, &status);
    proc = status.MPI_SOURCE;
    color = proc_color_.find(proc+1);
    sz = 0; neq = 0;
    while(pb-buff < tt) {
      todo--;                                                 assert(todo>=0);
      int gbid = *pb++, blockGID = *pb++; 
      neq += bs_;
      if( blockGID >= 0 ){
	const PromGSNode * const node = gbid_ghost_.find(gbid+1);
	assert(node!=NULL);
	int lid = node->b.lid_ + nbloc; 
	lid_blockGID[lid] = blockGID; // ghost in IS
	// use global block IDs with color
	if( !gbid_ISBlock_.find(blockGID + 1) ){
	  // create IS_BLOCK (lite wieght for ghosts)
	  PromGSBlock *blk = new PromGSBlock( blockGID, proc ); CHKNEW(blk,1);
	  int add = gbid_ISBlock_.add(blk, blockGID+1); assert(add>0);
	  assert( proc!= mype );
	}
      }
      else assert( blockGID == -1 ); // ????
    }
    assert(tt==pb-buff); 
    // make recvlists
    sz = tt/2; assert(sz>0); tg = (color < mycolor) ? hightag : lowtag;
    PromProcXComm *commx = new PromProcXComm( sz, neq, proc, comm_, tg );  
    CHKNEW(commx,1);
    for( ii = 0, tt = 0 ; ii < sz ; ii++, tt += 2 ){
      int gid = buff[tt];
      // indices -- translate into local
      PromGSNode *node = gbid_ghost_.find(gid+1); 
      if(node == NULL){
	PetscPrintf(PETSC_COMM_SELF,"[%d]%s gid=%d\n",mype,__FUNCT__,gid);
	SETERRQ1(1,"ERROR: missing node %d",gid);	
      }
      else { commx->lids_[ii] = node->b.lid_;   assert(node->b.lid_>=0); }
    }
    if( color < mycolor ) lowProc_recv_.addTail( commx );
    else highProc_recv_.addTail( commx );
    assert( commx->req_ == MPI_REQUEST_NULL);
    // post recv -- ready for solve
    MPI_Irecv( commx->xarr_, commx->nEqs_ * sizeof(double)/sizeof(int),
	       MPI_INT, commx->part_, commx->tag_,
	       commx->comm_, &commx->req_ ); // *1*
    //
    PetscFree(buff);
  }

  // cleanup sends
  while( !smessages.isEmpty() ) {
    MPI_Request *sreq = smessages.removeHead();
    MPI_Wait( sreq, &status );    PetscFree( sreq );
  } 

  PetscFunctionReturn(0); 
}

#if defined(PARCH_win32) || defined(PARCH_ascired)
extern "C" long random();
#endif

#undef __FUNCT__  
#define __FUNCT__ "PromPCGaussSeidel::MakeLists"
/* PromPCGaussSeidel::MakeLists ***********************************
 * 
 *   INPUT:
 *    - block_procs[nblocks]: processors that each block talks to (out)
 *    - block_gid_ghostLid[nblocks]: ghosts that each block talks to 
 *    - block_type[nblocks]: 
 *    - lid_blockGID[nbloc+nghosts]: 
 *    - block_nodeLids[nblocks]: lid of nodes in block
 *    - block_boundLids[nblocks]: lid of boundary nodes in block
 *    - block_adjacBLids[nblocks]: local block ids of adjacs of block
 *    - domainISs: 
 *
 *   SIDE EFFECTS:
 *    - fills in 'block->high(low)Proc_sends' and 'block->high(low)_ISBlocks'
 *    - fills in 'blocksLists_'
 *
 *   RETURN:
 *     - PETSc error code
 */
int PromPCGaussSeidel::MakeLists(PromList<int> * const * const block_procs,
				 const PromTable<int>*const*const
				 block_gid_ghostLid,
				 const PromBlockType block_type[], 
				 const int lid_blockGID[],
				 const PromList<int>*const*const
				 block_nodeLids,
				 const PromList<int>*const*const
				 block_boundLids, 
				 const PromList<int>*const*const
				 block_adjacBLids,
				 const PromIS * const domainISs )
{
  int    ierr,color,proc,haslowerIS,hashigherIS,gbid,sz,xx,nInt1,nnzTot,neq;
  int    proc2,jj,hasgh,ncols,adjacs[PROM_MAXNIEGH],infsMinusSups,nup,ndown;
  ierr = MatGetOwnershipRange( A_->mat_, &xx, &sz); CHKERRQ( ierr );
  const int mybnext0=sz/bs_,myb0=xx/bs_,nbloc = mybnext0 - myb0;
  long  *buff,*randIDs,*blockNNZ;
  const int np = A_->map_.comm_.np(),mype = A_->map_.comm_.mype();
  const int mycolor = proc_color_.find(mype+1);
  const int nghosts = gbid_ghost_.getCount(), nbNodes = nghosts + nbloc;
  TablePos tipos;
  Position  pos,pos2;
  PromBlockType btype;
  PromTable<PromProcXComm*>   proc_lowCommx, proc_highCommx;

  PetscFunctionBegin;
  ierr = PetscMalloc((2*nGSBlocks_+nbloc+1)*sizeof(long),&randIDs);
  CHKERRQ(ierr);
  blockNNZ = randIDs + nGSBlocks_;
  buff = blockNNZ + nGSBlocks_;                 assert(nGSBlocks_<=nbloc);
  // make caches
  pos = highProc_recv_.getHeadPosition(); 
  while( pos ){ 
    PromProcXComm *commx = highProc_recv_.getNext( pos );
    int add = proc_highCommx.add( commx, commx->part_+1 ); assert(add);
  }
  pos = lowProc_recv_.getHeadPosition(); 
  while( pos ){ 
    PromProcXComm *commx = lowProc_recv_.getNext( pos );
    int add = proc_lowCommx.add( commx, commx->part_+1 ); assert(add);
  }

  // form 'blocksLists_' and IS data  
  for( xx = 0 ; xx < nGSBlocks_ ; xx++ ) randIDs[xx] = -1;
  for(xx=infsMinusSups=nInt1=nnzTot = nup = ndown = 0 ; xx < nGSBlocks_; xx++){
    const int blidi = xx;
    // count nnz in block
    pos = block_nodeLids[blidi]->getHeadPosition();  sz = 0; 
    while( pos ){ 
      const int lidi = block_nodeLids[blidi]->getNext(pos) - 1;
      ierr = A_->GetLocal_constBlk_NodeRow(lidi, PROM_MAXNIEGH, ncols, adjacs);
      CHKERRQ(ierr);
      sz += ncols;
    }
    blockNNZ[blidi] = sz;  nnzTot += sz;
    
    // get type, and procs
    btype = block_type[blidi];
    if( !block_gid_ghostLid[blidi] ) assert(btype == PROM_INTER);
    // add to a list
    switch( btype ) {
    default: SETERRQ(1,"error");
    case PROM_INTER: // interior -- mix up for red-black like behavior
      nInt1 += sz; // count Ints -- put int INT1 for now
      blocksLists_[PROM_INTER1].list_->addTail( blidi+1 ); 
      break;
    case PROM_INF: 
      infsMinusSups += sz; // count Infs
    case PROM_SUP: // fallthrough
      if( btype == PROM_SUP ) infsMinusSups -= sz; 
      assert( block_procs[blidi] != NULL );
      if( block_procs[blidi]->getCount() == 1 ){
	proc = block_procs[blidi]->getHead() - 1;  
	color = proc_color_.find(proc+1); assert(color>0);
	PromProcXComm *commx;
	if(color<mycolor) commx = proc_lowCommx.find(proc+1);
	else commx = proc_highCommx.find(proc+1);
	assert(commx!=NULL);
	commx->InfBlocks_.list_->addTail(blidi+1);
	assert(commx->InfBlocks_.arr_size_ == 0);
      }
      else if(btype==PROM_INF)blocksLists_[PROM_INFL].list_->addTail( blidi+1 );
      else blocksLists_[PROM_SUPL].list_->addTail( blidi+1 );
      break;
    case PROM_INF_SUP:
      assert( block_gid_ghostLid[blidi] != NULL );      
      const int gbidi = mype*maxnGSBlocks_ + blidi;
      PromGSBlock *gsblki = gbid_ISBlock_.find(gbidi+1); 
      PromInfSupBlock *ISlocal = dynamic_cast<PromInfSupBlock*>(gsblki);
      assert(ISlocal != NULL); 
      // see if hig/low IS adjacs and set ghost high(low)_ISBlocks.
      haslowerIS = hashigherIS = FALSE;
      tipos = block_gid_ghostLid[blidi]->getHeadPosition(); 
      while( tipos ){ 
	int lid = block_gid_ghostLid[blidi]->getNext( tipos ) - 1 + nbloc;
	assert(lid >= nbloc && lid < nbNodes);	
	if( (gbid=lid_blockGID[lid]) >= 0 ) { // a ghost IS
	  PromGSBlock *b2 = gbid_ISBlock_.find(gbid+1);
	  assert(b2!=NULL);
	  proc = b2->getProc();
	  color = proc_color_.find(proc+1); assert(color!=mycolor && color>0);
	  // see if high/low IS and add ghost high(low)_ISBlocks
	  if( color > mycolor ){
	    hashigherIS = TRUE;
	    if( !ISlocal->high_ISBlocks_.find( b2 ) ) {
	      ISlocal->high_ISBlocks_.addTail( b2 );
	    }
	  }
	  else {
	    haslowerIS = TRUE;
	    if( !ISlocal->low_ISBlocks_.find( b2 ) ) {
	      ISlocal->low_ISBlocks_.addTail( b2 );
	    }
	  }
	  b2->mark();//ghost IS 'b2' is added to *a* RECV list, so !add latter
	}
	else assert(lid_blockGID[lid] == -1);
      }

      //put in sub list for PROM_INF_SUP decending order! (should color within)
      if( hashigherIS && haslowerIS ){ 
	// has both -- 'true' IS
	blocksLists_[PROM_INF_SUPL].list_->addHead( blidi+1 ); 
	randIDs[blidi] = random();
      }
      else if( hashigherIS ) {
	assert( !haslowerIS ); // last in IS chain
	blocksLists_[PROM_INF_SUP_WO_LOWERIS].list_->addHead( blidi+1 );
      }
      else if( haslowerIS ) { 
	assert(!hashigherIS); // first in IS chain
	blocksLists_[PROM_INF_SUP_WO_HIGHERIS].list_->addHead( blidi+1 ); 
      }
      else { // has neither higher or lower inf_sup -- split
	if( nup < ndown ){
	  nup += sz;
	  blocksLists_[PROM_INF_SUP_WO_LOWERIS].list_->addHead( blidi+1 );
	} 
	else {
	  ndown += sz;
	  blocksLists_[PROM_INF_SUP_WO_HIGHERIS].list_->addHead( blidi+1 );
	} 
      }
      // form block's send objects
      assert(block_procs[blidi] != NULL && block_boundLids[blidi] != NULL);
      pos = block_procs[blidi]->getHeadPosition();  assert(pos!=NULL);
      while( pos ) {
	proc = block_procs[blidi]->getNext( pos ) - 1;  
	color = proc_color_.find(proc+1); assert(color != mycolor && color>0);
	// put bound nodes in 'buff' to make communicator
	const PromList<int> *adjs = block_boundLids[blidi];
	pos2 = adjs->getHeadPosition();  neq = sz = 0;
	while( pos2 ){
	  const int lidi = adjs->getNext(pos2) - 1;
	  ierr=A_->GetLocal_constBlk_NodeRow(lidi,PROM_MAXNIEGH,ncols,adjacs);
	  CHKERRQ(ierr);
	  for( jj = 0, hasgh = 0 ; jj < ncols ; jj++ ) {
	    const int gidj = adjacs[jj];
	    if( gidj < myb0 || gidj >= mybnext0 ){ 
	      ierr = GetProc( gidj, proc2 ); CHKERRQ(ierr);	      
	      if( proc2 == proc ) { hasgh = 1; break; }
	    }
	  }
	  // this row, in the block, touches this proc  
	  if( hasgh ){
	    buff[sz++] = lidi;  neq += bs_;
	  }
	}
	assert(sz > 0);
	// PromBlockXComm
	PromBlockXComm *commx = new PromBlockXComm(sz,neq,proc,gbidi);
	CHKNEW(commx,1);
	if( color > mycolor ) ISlocal->highProc_sendXComm_.addTail( commx );
	else ISlocal->lowProc_sendXComm_.addTail( commx );
	for( int ii = 0 ; ii < sz ; ii++ ) commx->locBoundIDs_[ii] = buff[ii];
      }
      break;
    }  // switch
  }    // blidi
  if( np > 1 ) {
    // put local IS blocs in dep. lists high(low)_ISBlocks (use 'randIDs')
    for( xx = 0 ; xx < nGSBlocks_; xx++ ) {
      const int blidi = xx, iind = randIDs[blidi];
      if( iind != -1 && block_adjacBLids[blidi] != NULL){
        const int gbidi = mype*maxnGSBlocks_ + blidi;
        PromGSBlock *gsblki = gbid_ISBlock_.find(gbidi+1);
        PromInfSupBlock *ISlocal = dynamic_cast<PromInfSupBlock*>(gsblki);
        pos = block_adjacBLids[blidi]->getHeadPosition();
        while( pos ){
          const int blidj = block_adjacBLids[blidi]->getNext( pos ) - 1;
          int jind = randIDs[blidj];
          if( jind != -1 ) {
            gbid = mype * maxnGSBlocks_ + blidj; // PromInfSupBlock
            PromGSBlock *blckj=gbid_ISBlock_.find(gbid+1); assert(blckj!=NULL);
            if( jind > iind || (jind==iind && blidj > blidi) ){
	      ISlocal->high_ISBlocks_.addHead( blckj );
	    }
            else ISlocal->low_ISBlocks_.addHead( blckj ); 
          }
        }
      }
    }
    // split Int -> Int1 and Int2
    if( blocksLists_[PROM_INTER1].list_->getCount() > 1 ) { 
      // balance around IS: 2*(Int1 == Int1B) Inf == 2(Int2 == Int2B) + Sup
      PromList<int> *int1=blocksLists_[PROM_INTER1].list_;
      PromList<int> *int1b=blocksLists_[PROM_INTER1B].list_;
      PromList<int> *int2=blocksLists_[PROM_INTER2].list_;
      PromList<int> *int2b=blocksLists_[PROM_INTER2B].list_;
      // infsMinusSups *= 1; // heuristic as inf and sups seem to run slower
      int sztodo = (nInt1+infsMinusSups)/2, nInt2=0;
      bool use_b_lists = (maxnGSBlocks_ > 100) ? TRUE : FALSE;
#if defined(PROM_HAVE_METIS)
      //#define PROM_USE_METIS
#endif 
#if defined(PROM_USE_METIS)
      int nnodes = int1->getCount();
      if( sztodo <= nInt1/20 || sztodo >= (19*nInt1)/20 || nnodes < 4 ) { 
	// small grid
	if( sztodo >= (19*nInt1)/20 ) { // just move into Int2
	  while( !int1->isEmpty() ) int2->addTail( int1->removeHead() );
	  nInt2 = nInt1 ; nInt1 = 0;
	  if( use_b_lists ) { 	        // dumb balance 
	    for( sz = 0 ; sz < nInt2/2 && !int2->isEmpty() ; ){	    // int2
	      int blidi = int2->removeHead() - 1; 
	      sz += blockNNZ[blidi]; 
	      int2b->addTail(blidi+1);
	    }
	  }
	}
	else if( use_b_lists ) { 	// dumb balance 
	  // int1
	  for( sz = 0 ; sz < nInt1/2 && !int1->isEmpty() ; ){
	    int blidi = int1->removeHead() - 1; 
	    sz += blockNNZ[blidi]; 
	    int1b->addTail(blidi+1);
	  }
	}
      }
      else {
	// set 'xadj' & 'adjncy' adjacentcy data
	idxtype *xadj, *adjncy, *nwts = NULL, *id_part;
	int numadjac,ii,npp;
	ierr = PetscMalloc(2*(nnodes+1)*sizeof(idxtype),&xadj); CHKERRQ(ierr);
	id_part = xadj + 1*(nnodes + 1); 
	//nwts = xadj + 2*(nnodes + 1); 
	for(ii=0;ii<nGSBlocks_;ii++) randIDs[ii] = -1;//use 'randIDs' as buffer
	for( pos = int1->getHeadPosition(), ii = 0 ; pos ; ii++ ){ 
	  int blidi = int1->getNext( pos ) - 1; assert(randIDs[blidi] == -1);
	  randIDs[blidi] = ii; // bid --> index
	  if(nwts != NULL) nwts[ii] = blockNNZ[blidi];
	}
	for( pos = int1->getHeadPosition(), ii = numadjac = 0 ; pos ; ii++ ){ 
	  int blidi = int1->getNext( pos ) - 1;
	  xadj[ii] = numadjac;
	  const PromList<int> *const adjs = block_adjacBLids[blidi];
	  if( adjs != NULL ) { // can happen with mixed dof (?)
	    pos2 = adjs->getHeadPosition();  
	    while( pos2 ){ 
	      int blidj = adjs->getNext( pos2 ) - 1, id = randIDs[blidj];
	      assert(blidj!=blidi); 
	      if( id != -1 ) numadjac++;
	    }
	  }
	}
	xadj[ii] = numadjac;  assert(ii == nnodes);
	if( numadjac <= nnodes ) {
	  // some multi-bodies
	  while( nInt2 < sztodo && !int1->isEmpty() ) {
	    int blidi = int1->removeHead() - 1;  sz = blockNNZ[blidi];
	    nInt2 += sz; nInt1 -= sz; // current size if int1
	    int2->addTail(blidi+1);  
	  }	  
	}
	else { // METIS
	  ierr = PetscMalloc(numadjac*sizeof(idxtype),&adjncy); CHKERRQ(ierr);
	  for( pos = int1->getHeadPosition(), ii = 0 ; pos ; /* void */ ){ 
	    int blidi = int1->getNext( pos ) - 1;
	    const PromList<int> *const adjs = block_adjacBLids[blidi];
	    if( adjs != NULL ) {
	      pos2 = adjs->getHeadPosition();  
	      while( pos2 ){ 
		int blidj = adjs->getNext( pos2 ) - 1, id = randIDs[blidj];
		if( id != -1 ) adjncy[ii++] = id;
	      }
	    }
	  }
	  assert(ii == numadjac);
	  // part.
	  float ft = (float)nInt1, ubvec=1.05; MPI_Comm scomm = MPI_COMM_SELF;
	  float tpwght[4]={(float)(nInt1-sztodo)/ft, (float)sztodo/ft, 0., 0.};
	  if( use_b_lists ){ 
	    npp = 4;
	    tpwght[2] = tpwght[3] = tpwght[1] / 2.; 
	    tpwght[0] = tpwght[1] = tpwght[0] / 2.; 
	  }
	  else npp = 2;
	  int options[5]={0,3,1,1,0},wflag=(nwts!=NULL)?2:0,volume,numf=0,ncon=1;
	  //METIS_WPartGraphRecursive(&nnodes,xadj,adjncy, nwts, NULL, &wflag,
	  // &numf,&npp,tpwght,options,&volume,id_part);
	  ParMETIS_V3_PartKway( &nnodes, xadj, adjncy, nwts, NULL, &wflag, 
				&numf, &ncon, &npp, tpwght, &ubvec, options, &volume, 
				id_part, &scomm );
	  // do lists
	  for(ii = 0 ; ii < nnodes ; ii++){ 
	    int blidi=int1->removeHead()-1,id=randIDs[blidi],pid=id_part[id]; 
	    if( pid < npp/2 ) { // int 1
	      if( pid == 0 ) int1->addTail(blidi+1);
	      else int1b->addTail(blidi+1);
	    }
	    else {	  // int2
	      sz = blockNNZ[blidi];
	      nInt2 += sz;  nInt1 -= sz;
	      if( pid == 3 ) int2b->addTail(blidi+1);
	      else int2->addTail(blidi+1);
	    }
	  }
	  PetscFree(adjncy);
	}
	PetscFree(xadj);   
      }
#else
      for(xx=0;xx<nGSBlocks_;xx++) randIDs[xx] = 0;//use 'randIDs' as buffer
      for( pos = int1->getHeadPosition() ; pos ; /*void*/ ){ 
	int blidi = int1->getNext( pos ) - 1; assert(randIDs[blidi] == 0);
	randIDs[blidi] = 1; // here 
      }
      PromList<int> nextRoots(20); 
      int blidi = int1->removeHead() - 1;  nextRoots.addTail(blidi+1);
      int sz = blockNNZ[blidi]; nInt2 += sz; nInt1 -= sz; 
      int2->addTail(blidi+1); randIDs[blidi] = 2; 
      while(1){
	while( !nextRoots.isEmpty() && nInt2 < sztodo && !int1->isEmpty() ) {
	  blidi = nextRoots.removeHead()-1; assert(randIDs[blidi] == 2);
	  if( block_adjacBLids[blidi] != NULL ) {
	    const PromList<int>*const adj = block_adjacBLids[blidi];
	    pos = adj->getHeadPosition(); 
	    while( pos ){ 
	      blidi = adj->getNext( pos ) - 1;
	      if( randIDs[blidi] == 1 ){
		sz = blockNNZ[blidi]; nInt2 += sz; nInt1 -= sz; 
		int2->addTail(blidi+1); randIDs[blidi] = 2;
		nextRoots.addTail(blidi+1); 
		if(nInt2 >= sztodo || int1->isEmpty()) break;
	      }
	    }
	  }
	}
	if( nInt2 >= sztodo || int1->isEmpty() ) break;
	// get next root
	assert(nextRoots.isEmpty());
	for( pos = int1->getHeadPosition() ; pos ; /*void*/ ){ 
	  blidi = int1->getNext( pos ) - 1; 
	  if( randIDs[blidi] == 1 ){
	    sz = blockNNZ[blidi]; nInt2 += sz; nInt1 -= sz; 
	    int2->addTail(blidi+1); randIDs[blidi] = 2;
	    nextRoots.addTail(blidi+1); 
	    break;
	  }
	}
	if( nextRoots.isEmpty() ) break;
      }
      // balance 
      if( use_b_lists ) {
	// int1
	for( sz = 0 ; sz < nInt1/2 && !int1->isEmpty() ; ){
	  int blidi = int1->removeHead() - 1; 
	  sz += blockNNZ[blidi]; 
	  int1b->addTail(blidi+1);
	}
	// int2
	for( sz = 0 ; sz < nInt2/2 && !int2->isEmpty() ; ){
	  int blidi = int2->removeHead() - 1; 
	  sz += blockNNZ[blidi]; 
	  int2b->addTail(blidi+1);
	}
      }
#endif
    }
  } // if ( np > 1 )
  
  { // reorder inf and sup lists
    PromList<int> list(100),nextRoots; 
    bool bf1__[PROM_BSZ], bf2__[PROM_BSZ], *bf1, *bf2; 
    if( nGSBlocks_ >= PROM_BSZ ) {
      ierr = PetscMalloc(nGSBlocks_*sizeof(int),&bf1); CHKERRQ(ierr);  
      ierr = PetscMalloc(nGSBlocks_*sizeof(int),&bf2); CHKERRQ(ierr);  
    }
    else{  bf1 = bf1__; bf2 = bf2__; }
    for( xx = 0 ; xx < nGSBlocks_ ; xx++ ) bf1[xx] = bf2[xx] = FALSE;  
    ierr = BfsReorder( blocksLists_[PROM_INTER1].list_, block_adjacBLids,
		       &list, &nextRoots, bf1, bf2, options_.verbose_>2 ); 
    CHKERRQ(ierr);
    pos = highProc_recv_.getHeadPosition();
    while( pos ){
      PromProcXComm *commx = highProc_recv_.getNext( pos );
      ierr = BfsReorder( commx->InfBlocks_.list_, block_adjacBLids, &list,
			 &nextRoots, bf1, bf2 ); 
      CHKERRQ(ierr);
    }
    pos = lowProc_recv_.getHeadPosition();
    while( pos ){
      PromProcXComm *commx = lowProc_recv_.getNext( pos );
      ierr = BfsReorder( commx->InfBlocks_.list_, block_adjacBLids, &list,
			 &nextRoots, bf1, bf2 );
      CHKERRQ(ierr);
    }
    ierr = BfsReorder( blocksLists_[PROM_INFL].list_, block_adjacBLids,
		       &list, &nextRoots, bf1, bf2 );
    CHKERRQ(ierr);
    ierr = BfsReorder( blocksLists_[PROM_SUPL].list_, block_adjacBLids,
		       &list, &nextRoots, bf1, bf2 );
    CHKERRQ(ierr);
    if( bf1 != bf1__ ){ PetscFree(bf1); PetscFree(bf2); }
  }

  // clear marks
  TablePos tpos = gbid_ISBlock_.getHeadPosition(); 
  while( tpos ){ 
    PromGSBlock *block = gbid_ISBlock_.getNext( tpos ); 
    block->mark(TRUE); // all done w/marks - now as if it was just used
  }
  PetscFree(randIDs);  
  PetscFunctionReturn(0); 
}
  
#undef __FUNCT__  
#define __FUNCT__ "PromPCGaussSeidel::BfsReorder"
/* PromPCGaussSeidel::BfsReorder *******************************************
 *
 *   - reorder list (ie,Cuthill-McKee)
 *
 *   INPUT:
 *    - blidList: 'node' target list (in/out)
 *    - block_adjBLids: 'edges' 
 *    - list: temp list
 *    - nextRoots: temp list
 *    - donetab: 
 *    - heretab:
 *    - print:
 *
 *   SIDE EFFECTS:
 *   RETURN:
 *     - PETSc error code*
 ***********************************************************************/
int PromPCGaussSeidel::BfsReorder( PromList<int> *blidList, 
				   const PromList<int>*const*const
				   block_adjBLids, 
				   PromList<int> *list,
				   PromList<int> *nextRoots,
				   bool donetab[], bool heretab[],
				   const bool print/*=false*/ )
{
  assert(blidList != NULL);
  if( blidList->isEmpty() ) return 0;
  
  int kk,jj,rootblidi,minadj,ndone = 0, nn = blidList->getCount();
  Position pos,pos2;
  PetscFunctionBegin;
  // get rootblidi, init table
  pos = blidList->getHeadPosition();
  while(pos){
    jj = blidList->getNext(pos)-1; heretab[jj] = TRUE; assert(!donetab[jj]);
  }
  pos = blidList->getHeadPosition();  minadj = 1000000; rootblidi = -1;
  while( pos){
    jj = blidList->getNext( pos ) - 1; 
    if( block_adjBLids[jj] == NULL ){ minadj = 0; rootblidi = jj; }
    else{
      pos2 = block_adjBLids[jj]->getHeadPosition(); int adj = 0;
      while( pos2 ){ 
	kk = block_adjBLids[jj]->getNext( pos2 ) - 1;
	if( heretab[kk] == TRUE ) adj++;
      }
      if( adj < minadj ) { minadj = adj; rootblidi = jj; }
    }
  }
  assert(rootblidi != -1);
  // add root
  assert( list->isEmpty() && nextRoots->isEmpty());
  jj = rootblidi;
  donetab[jj] = TRUE; list->addTail(jj+1); nextRoots->addTail(jj+1); ndone++;
  while( ndone < nn ) {
    rootblidi = nextRoots->removeHead() - 1;
    if( block_adjBLids[rootblidi] != NULL ) {
      // add al of roots adjacs to list
      pos = block_adjBLids[rootblidi]->getHeadPosition(); 
      while( pos ){ 
	jj = block_adjBLids[rootblidi]->getNext( pos ) - 1;
	if( heretab[jj] && !donetab[jj] ){
	  donetab[jj] = TRUE; list->addTail(jj+1); 
	  nextRoots->addTail(jj+1); ndone++;
	}
      }
    }
    if( ndone < nn && nextRoots->isEmpty() ) {
      if( print ) // wrong!
	PetscPrintf( PETSC_COMM_WORLD,
		     "\t[0]%s WARNING Found disconnected in list (%d/%d)\n",
		     __FUNCT__,ndone,nn);
      // get next root
      assert( !blidList->isEmpty() );
      // get rootblidi, init table
      pos = blidList->getHeadPosition();  minadj = 1000000; rootblidi = -1;
      while(pos){
	jj = blidList->getNext(pos)-1; 
	if( !donetab[jj] ) {
	  if( block_adjBLids[jj] == NULL ){ minadj = 0; rootblidi = jj; }
	  else{
	    pos2 = block_adjBLids[jj]->getHeadPosition(); int adj = 0;
	    while( pos2 ){ 
	      kk = block_adjBLids[jj]->getNext( pos2 ) - 1;
	      if( heretab[kk] == TRUE && !donetab[kk] ) adj++;
	    }
	    if( adj < minadj ) { minadj = adj; rootblidi = jj; }
	  }  
	}
      }
      assert(rootblidi != -1);
      // add root
      jj = rootblidi;
      donetab[jj] = TRUE; list->addTail(jj+1); nextRoots->addTail(jj+1);
      ndone++;
    }
  }
  assert( ndone == nn );
  while( !blidList->isEmpty() ) blidList->removeHead();   // remove rest
  while( !nextRoots->isEmpty() ) nextRoots->removeHead(); // clear buffers
  while( !list->isEmpty() ) {
    jj = list->removeHead() - 1;
    blidList->addTail(jj+1); // reorder
    donetab[jj] = FALSE; heretab[jj] = FALSE;
  }
  PetscFunctionReturn(0); 
}
 
#undef __FUNCT__  
#define __FUNCT__ " PromPCGaussSeidel::Apply_private"
/* PromPCGaussSeidel::Apply_private ****************************************
 * 
 * - used for one level solves only
 *
 */     
int PromPCGaussSeidel::Apply_private( const PromVector_base * const BB,
				      PromVector_base * const XX,
				      const bool zeroguess /*=TRUE*/ )
{
  int ierr, itmax = numits_; // (symmetric_) ? 2 * numits_ : numits_;

  PromVector *xx = dynamic_cast<PromVector*>(XX);             assert(xx!=NULL);
  const PromVector *bb = dynamic_cast<const PromVector*>(BB); assert(bb!=NULL);
  // doit
  perf_mon_.EventBegin(PromContext::prom_logs_[LOG_MIS3]); 
  ierr = Apply_one( bb, xx, zeroguess, !zeroguess  ); CHKERRQ(ierr);
  if(symmetric_) pre_smooth_ = !pre_smooth_; //toggle -- hack for up/down
  for( int it = 1 ; it < itmax ; it++ ) {
    ierr = Apply_one( bb, xx, FALSE, FALSE  ); CHKERRQ(ierr);
    if(symmetric_) pre_smooth_ = !pre_smooth_; // toggle -- hack for up/down
  }
  perf_mon_.EventEnd(PromContext::prom_logs_[LOG_MIS3]);
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromPCGaussSeidel::Apply_one"
/* PromPCGaussSeidel::Apply_one ***********************************
 *
 *   - Gauss-Seidel smoother (collective)
 *
 *   INPUT:
 *    - bb: rhs
 *    - xx: solve (has initail guess in it!)
 *    - zeroguess
 *    - send_post: flag to send/recv and post for next
 *
 *   SIDE EFFECTS:
 *    - reorders 'lowProc_recv_' or 'highProc_recv_'
 *
 *   RETURN:
 *     - PETSc error code
 */  
int PromPCGaussSeidel::Apply_one( const PromVector * const bb, 
				  PromVector * const xx, 
				  bool zeroguess /*=FALSE*/, 
				  bool send_post /*=TRUE*/ ) 
{
  int         ierr,pre_s=pre_smooth_,proc,nrecv,nn,jj;
  const int   mype = A_->map_.comm_.mype(),IStag = A_->map_.comm_.getNewTag(comm_);
  const int   mycolor = proc_color_.find(mype+1);
  PromGSCont  *inf, *int1, *int1b, *IS_wo_hiIS, *InfSup;
  PromGSCont  *IS_wo_lwIS, *int2, *int2b, *sup;
  const PromList<PromProcXComm*> *lowPr_s, *highPr_s;
  PromAList<PromProcXComm*> *highPr_r, *lowPr_r; 
  TablePos    tpos;
  MPI_Status  status; 
  Position    pos,pos2; PromList<MPI_Request*> smessages;
  PromList<PromBlockXComm*> *list;
  PromTable<PromList<PromBlockXComm*>*> proc_list(57);

  if( comm_ == MPI_COMM_NULL ) return 0; // active only
  const int nghostEqs = gbid_ghost_.getCount() * bs_; 

  // get pointers to block lists (for symmetry)
  PetscFunctionBegin;
  if( pre_s ){
    lowPr_s = &lowProc_send_;    lowPr_r = &lowProc_recv_;
    highPr_r = &highProc_recv_;    highPr_s = &highProc_send_; 
    inf = &blocksLists_[PROM_INFL];
    int1b =  &blocksLists_[PROM_INTER1B];
    int1 =  &blocksLists_[PROM_INTER1]; 
    IS_wo_hiIS = &blocksLists_[PROM_INF_SUP_WO_HIGHERIS];
    InfSup = &blocksLists_[PROM_INF_SUPL];
    IS_wo_lwIS = &blocksLists_[PROM_INF_SUP_WO_LOWERIS];
    int2 = &blocksLists_[PROM_INTER2];
    int2b = &blocksLists_[PROM_INTER2B]; 
    sup = &blocksLists_[PROM_SUPL];
  } 
  else {
    lowPr_s = &highProc_send_;    lowPr_r = &highProc_recv_;
    highPr_r = &lowProc_recv_;    highPr_s = &lowProc_send_; 
    inf = &blocksLists_[PROM_SUPL];
    int1b =  &blocksLists_[PROM_INTER2B]; 
    int1 =  &blocksLists_[PROM_INTER2]; 
    IS_wo_hiIS = &blocksLists_[PROM_INF_SUP_WO_LOWERIS];
    InfSup = &blocksLists_[PROM_INF_SUPL];
    IS_wo_lwIS = &blocksLists_[PROM_INF_SUP_WO_HIGHERIS];
    int2 = &blocksLists_[PROM_INTER1]; 
    int2b = &blocksLists_[PROM_INTER1B]; 
    sup = &blocksLists_[PROM_INFL];
  }
  jj = nghostEqs; assert(ghostx_[jj] == -12.);
  if( zeroguess ){
    while(jj--) ghostx_[jj] = 0.0;
    ierr = xx->Set( 0.0 ); CHKERRQ(ierr); assert(send_post==FALSE);
  }
  
  // count recvs
  int IStoRecv = 0, IStoSend = 0;
  // init IS blocks, count sends and recvs
  tpos = gbid_ISBlock_.getHeadPosition(); 
  while( tpos ){ 
    PromGSBlock *blck = gbid_ISBlock_.getNext( tpos ); 
    assert(blck->isMarked());                              blck->mark( FALSE );
    proc = blck->getProc();
    if( proc == mype ) {
      PromInfSupBlock *ISlocal = dynamic_cast<PromInfSupBlock*>(blck);
      assert(ISlocal != NULL);
      if( pre_s ) {
	ISlocal->ISBlockDeps_ = &ISlocal->high_ISBlocks_;
	ISlocal->sendXComm_ = &ISlocal->lowProc_sendXComm_;
      } 
      else {
	ISlocal->ISBlockDeps_ = &ISlocal->low_ISBlocks_;
	ISlocal->sendXComm_ = &ISlocal->highProc_sendXComm_;
      }
      IStoSend++; // will send block->highProc_sends->getCount() 
    }
    else{ // ghost IS: recv from higher ISs
      int color = proc_color_.find(proc+1); 
      if( pre_s && color > mycolor ) IStoRecv++;  
      else if( !pre_s && color < mycolor ) IStoRecv++; 
      else blck->mark(); // mark as done as will not use it ('lower' IS block)
    } 
#if defined(PROM_MEASURE_MPATH)
    blck->path_ = 0;
#endif
  }
  // start
#if defined(PROM_USE_VAMPIRE) 
  VT_begin(1);   
#endif
  // send all to higher *1*
  if( send_post ) {
    ierr = SendLocalx( highPr_s, xx ); CHKERRQ(ierr);
  } 
  // do interior 1 blocks
  ierr = DoBlocks( int1b->arr_size_, int1b->array_, bb, xx );  CHKERRQ(ierr);
#if defined(PROM_USE_VAMPIRE) 
  VT_end(1);  VT_begin(2);
#endif

  perf_mon_.EventBegin(PromContext::prom_logs_[LOG_MIS4]);
  // recv all from lower *1* -- do infs w/ one proc
  if( send_post ) {
    ierr = RecvDoGhostx( lowPr_r, bb, xx ); CHKERRQ(ierr);
  }
  else {
    // do inf w/ 1 proc
    lowPr_r->count_++;
    pos = lowPr_r->getHeadPosition(); 
    while( pos ){
      PromProcXComm *commx = lowPr_r->getNext( pos );
      ierr = DoBlocks( commx->InfBlocks_.arr_size_, 
		       commx->InfBlocks_.array_, bb, xx); 
      CHKERRQ(ierr);
    }
  }
  // do rest of inf blocks 
  ierr = DoBlocks( inf->arr_size_, inf->array_, bb, xx );  CHKERRQ(ierr);
  perf_mon_.EventEnd(PromContext::prom_logs_[LOG_MIS4]);

#if defined(PROM_USE_VAMPIRE) 
  VT_end(2); VT_begin(3);
#endif
  // send all to lower *2*
  ierr = SendLocalx( lowPr_s, xx );  CHKERRQ(ierr);
  // do interior 1 blocks
  ierr = DoBlocks( int1->arr_size_, int1->array_, bb, xx );  CHKERRQ(ierr);
#if defined(PROM_USE_VAMPIRE) 
  VT_end(3);
#endif
  const bool noIS = ( IS_wo_hiIS->arr_size_==0 && InfSup->arr_size_==0 &&
		      IS_wo_lwIS->arr_size_==0 && 0 );
  if( !noIS ){ // !(no IS nodes (lowest or highest color))
#if defined(PROM_USE_VAMPIRE) 
    VT_begin(4);
#endif 
    // recv all from higher *2*
    ierr = RecvDoGhostx( highPr_r );   CHKERRQ(ierr);
    
    // do inf_sup blocks -- w/o higher inf_sup (but higher inf)
    ierr = DoBlocks( IS_wo_hiIS->arr_size_, IS_wo_hiIS->array_, bb, xx );
    CHKERRQ(ierr);
    // SEND IS_wo_hiIS
    for( jj = 0 ; jj < IS_wo_hiIS->arr_size_ ; jj++ ) { 
      int blid = IS_wo_hiIS->array_[jj], gbid = mype * maxnGSBlocks_ + blid;
      PromGSBlock *isb = gbid_ISBlock_.find(gbid+1); assert(isb!=NULL);
      PromInfSupBlock *ISlocal = dynamic_cast<PromInfSupBlock*>(isb);
      assert(!ISlocal->isMarked()); ISlocal->mark(); // done
      IStoSend--;
      ierr = ISlocal->collectSends( proc_list ); CHKERRQ(ierr);
    }
    // isend
    ierr = SendISblocks( &proc_list, &smessages, xx, IStag ); CHKERRQ(ierr);
    
    // IS loop
    int todo = InfSup->arr_size_; PromList<int> tlist(10); 
    while( todo != 0 ) {
      // recv from higher inf_sup *3*
      nrecv = 0;
      while( IStoRecv ) { 
	ierr = RecvBlockGhostx(ghostx_, nghostEqs, IStag, (nrecv>0), nn);
	CHKERRQ(ierr);
	if( nn == 0 ) break;
	IStoRecv -= nn;                  assert(IStoRecv >= 0);
	nrecv += nn;  	                                  
      }
      // try to do any IS block -- iterate in symmetric order
      assert(tlist.isEmpty());
      while( 1 ){
	// iterate 'InfSup' 
	int done=0,k0=(pre_s ? 0 : InfSup->arr_size_-1),kinc=(pre_s ? 1 : -1);
	for( int kk = 0 ; kk < InfSup->arr_size_ ; kk++, k0 += kinc ){
	  const int blid = InfSup->array_[k0];
	  int gbid = mype*maxnGSBlocks_ + blid;
	  PromGSBlock *isb = gbid_ISBlock_.find(gbid+1);
	  PromInfSupBlock *ISlocal = dynamic_cast<PromInfSupBlock*>(isb);
	  assert(ISlocal!=NULL);
	  if( !ISlocal->isMarked() ) {
	    // see if all deps are done
	    int deps = 0;
	    pos2 = ISlocal->ISBlockDeps_->getHeadPosition(); 
	    while( pos2 && deps == 0 ){ 
	      PromGSBlock *b2 = ISlocal->ISBlockDeps_->getNext(pos2);
	      if( !b2->isMarked() ) deps++;
#if defined(PROM_MEASURE_MPATH)
	      if(b2->path_ >= ISlocal->path_)ISlocal->path_ = b2->path_+1;
#endif
	    } 
	    if( deps == 0 ) { 
	      int blid = ISlocal->getGBID() - mype * maxnGSBlocks_;
	      if( pre_s ) tlist.addTail(blid+1); // must match 'DoBlocks'
	      else tlist.addHead(blid+1); //add into temp list in orig. order!
	      ISlocal->mark();            // mark as done here to free others!
	      done++;                     // done one more
	    }
	  }
	}
	if( done == 0 ) break; // do until no more can be done
      }
      // send done
      if( !tlist.isEmpty() ) {
	// doit
	int buff__[PROM_BSZ], *buff, sz = tlist.getCount();
	if( sz > PROM_BSZ ) {
	  ierr = PetscMalloc(sz*sizeof(int),&buff); CHKERRQ(ierr);  
	}
	else buff = buff__;
	Position pos = tlist.getHeadPosition(); sz = 0;
	while( pos ) buff[sz++] = tlist.getNext(pos) - 1; 
	ierr = DoBlocks( sz, buff, bb, xx ); CHKERRQ(ierr);
	if( buff != buff__ ) PetscFree(buff);  
	while( !tlist.isEmpty() ) {
	  int blid = tlist.removeHead() - 1, gbid = mype*maxnGSBlocks_+blid;
	  PromGSBlock *isb = gbid_ISBlock_.find(gbid+1);
	  PromInfSupBlock *ISlocal = dynamic_cast<PromInfSupBlock*>(isb);
	  IStoSend--; todo--;  assert(todo>=0);
	  ierr = ISlocal->collectSends( proc_list ); CHKERRQ(ierr);
	}
	// isend
	ierr = SendISblocks( &proc_list, &smessages, xx, IStag );
	CHKERRQ(ierr);
      }
    } // outer IS loop

    // do inf_sup blocks -- w/o lower inf_sup (but lower inf)
    while( IStoRecv ) { 
      int nn;
      ierr = RecvBlockGhostx( ghostx_, nghostEqs, IStag, FALSE, nn );
      CHKERRQ(ierr);
      IStoRecv -= nn;                  assert(IStoRecv >= 0);
    }
    // do inf_sup blocks
    ierr = DoBlocks( IS_wo_lwIS->arr_size_, IS_wo_lwIS->array_, bb, xx );
    CHKERRQ(ierr);
    // SEND IS_wo_lwIS
    for( jj = 0 ; jj < IS_wo_lwIS->arr_size_ ; jj++ ) { 
      int blid = IS_wo_lwIS->array_[jj], gbid = mype * maxnGSBlocks_ + blid;
      PromGSBlock *isb = gbid_ISBlock_.find(gbid+1); assert(isb!=NULL);
      PromInfSupBlock *ISlocal = dynamic_cast<PromInfSupBlock*>(isb);
      assert(!ISlocal->isMarked()); ISlocal->mark(); // done
      IStoSend--;
      ierr = ISlocal->collectSends( proc_list ); CHKERRQ(ierr);
    }
    // isend
    ierr = SendISblocks( &proc_list, &smessages, xx, IStag ); 
    CHKERRQ(ierr);
    
#if defined(PROM_USE_VAMPIRE) 
    VT_end(4);
#endif  
  }
#if defined(PROM_USE_VAMPIRE) 
  VT_begin(5);
#endif
  
  // do interior 2b blocks
  ierr = DoBlocks( int2->arr_size_, int2->array_, bb, xx );   CHKERRQ(ierr);
  if( noIS ){
    // recv all from higher *2*
    ierr = RecvDoGhostx( highPr_r );    CHKERRQ(ierr);
  }
  // recv all rest (sups) -- recv from higher inf_sup *3*
  while( IStoRecv ) {
    ierr = RecvBlockGhostx( ghostx_, nghostEqs, IStag, FALSE, nn );
    CHKERRQ( ierr );
    IStoRecv -= nn;                    assert(IStoRecv >= 0);
  }
#if defined(PROM_USE_VAMPIRE) 
  VT_end(5);   VT_begin(6);
#endif
  perf_mon_.EventBegin(PromContext::prom_logs_[LOG_MIS5]);
  // do rest of sup blocks
  ierr = DoBlocks( sup->arr_size_, sup->array_, bb, xx );  CHKERRQ(ierr);
  // do sup w/ 1 proc
  highPr_r->count_++;
  pos = highPr_r->getTailPosition(); 
  while( pos ){
    PromProcXComm *commx = highPr_r->getPrev( pos );
    ierr = DoBlocks( commx->InfBlocks_.arr_size_, 
		     commx->InfBlocks_.array_, bb, xx ); 
    CHKERRQ(ierr);
  }
  perf_mon_.EventEnd(PromContext::prom_logs_[LOG_MIS5]);
#if defined(PROM_USE_VAMPIRE) 
  VT_end(6);   VT_begin(7);
#endif
  // do interior 2b blocks
  ierr = DoBlocks( int2b->arr_size_, int2b->array_, bb, xx );   CHKERRQ(ierr);
  while( !proc_list.isEmpty() ) {
    list = proc_list.removeHead();
    assert(list->isEmpty() );
    delete list; 
  }
  if( send_post ) {
    ierr = WaitForList( highPr_s ); CHKERRQ(ierr);// *1* 
  }
  ierr = WaitForList( lowPr_s ); CHKERRQ(ierr);// *2* 
  // clean up
  while( !smessages.isEmpty() ) {
    MPI_Request *sreq = smessages.removeHead();
    MPI_Wait( sreq, &status );    PetscFree( sreq );
  }
  // get ready for next -- post recv all from lower *1*
  if( send_post ) {
    ierr = PostRecvs( lowPr_r ); CHKERRQ(ierr);
  }

  // post recv all from higher *2*
  ierr = PostRecvs( highPr_r ); CHKERRQ(ierr);
#if defined(PROM_USE_VAMPIRE) 
  VT_end(7); 
#endif
#if defined(PROM_MEASURE_MPATH)
  static int vt_ndone = 0; 
  if( vt_ndone++ == 0 ){ 
    int maxpath = 0;
    tpos = gbid_ISBlock_.getHeadPosition();  
    while( tpos ){ 
      PromGSBlock *blck = gbid_ISBlock_.getNext( tpos ); 
      if( blck->path_ > maxpath ) maxpath = blck->path_;
    }
    MPI_Allreduce( &maxpath, &nn, 1, MPI_INT, MPI_MAX, comm_ );
    PetscPrintf(comm_,"\t\t\t[%d] maxpath=%d (%d)\n",mype,nn,maxpath);
  }
#endif
  
  PetscFunctionReturn(0); 
}

#undef __FUNCT__  
#define __FUNCT__ "PromInfSupBlock::collectSends"
/* PromInfSupBlock::collectSends ***********************************
 *
 *   INPUT: 
 *    - proc_list: send data added here
 *
 *   SIDE EFFECTS:
 *    - can add list to proc_list
 *
 *   RETURN:
 *     - PETSc error code
 */
int PromInfSupBlock::collectSends( PromTable<PromList<PromBlockXComm*>*> 
				   &proc_list )
{
  // colect sends 
  PromList<PromBlockXComm*> *list;
  Position pos2 = sendXComm_->getHeadPosition(); 

  while( pos2 ){
    PromBlockXComm *commx = sendXComm_->getNext(pos2);
    list = proc_list.find(commx->part_+1);
    if( list == NULL ) {
      list = new PromList<PromBlockXComm*>(12); CHKNEW(list,1);
      proc_list.add( list, commx->part_+1 );
    }
    list->addTail( commx );
#if defined(PROM_MEASURE_MPATH)
    commx->path_ = path_;
    assert(path_>=0);
#endif
  }
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromPCGaussSeidel::RecvBlockGhostx"
/* PromPCGaussSeidel::RecvBlockGhostx ***********************************
 *
 *   INPUT: 
 *    - ghostx[nghostEqs]: (out)
 *    - nghostEqs:
 *    - tag
 *    - iprobe
 *    - ndone (out)
 *
 *   SIDE EFFECTS:
 *
 *   RETURN:
 *     - PETSc error code
 */
int PromPCGaussSeidel::RecvBlockGhostx( double ghostx[], const int nghostEqs,
					const int tag, const bool iprobe,
					int &ndone ) const
{
  int ierr,ii,sz,*pb,*pb0,gbid,count,jj,recv,*buffer=NULL,bsz=0;
  MPI_Status status; const int fisz = sizeof(double)/sizeof(int);
  PromGSBlock *blck;
  int __rb[PROM_BSZ+1], *rbuf=__rb + ((size_t)__rb%sizeof(double)==0 ? 0 : 1);
  
  ndone = 0;
  while(1){ 
    if(iprobe || ndone > 0) MPI_Iprobe(MPI_ANY_SOURCE,tag,comm_,&recv,&status);
    else { MPI_Probe( MPI_ANY_SOURCE, tag, comm_, &status ); recv = TRUE; }
    if( !recv ) break;
    MPI_Get_count( &status, MPI_INT, &sz ); 
    if( sz < PROM_BSZ ) pb0 = pb = rbuf;
    else if( sz < bsz ) pb0 = pb = buffer;
    else {
      bsz = 2*sz;
      if( buffer != NULL ) PetscFree(buffer);
      ierr = PetscMalloc(bsz*sizeof(int),&buffer); CHKERRQ(ierr);
      pb0 = pb = buffer;
    }
    assert( (size_t)pb % sizeof(double) == 0 );
    
    // recv(tag) *< gbid, nn, nn*<gid, [-2,] bs*<val> > >
    MPI_Recv(pb0,sz,MPI_INT,status.MPI_SOURCE,tag,comm_,&status);
    while( pb-pb0 < sz ) {
      gbid = *pb++; count = *pb++;  
      blck = gbid_ISBlock_.find(gbid+1);      assert(blck!=NULL);
      ndone++; assert( !blck->isMarked() ); blck->mark();  
      // add in x update
      for( ii = 0 ; ii < count ; ii++ ) {
	int gid = *pb++;
	const PromGSNode * const node = gbid_ghost_.find(gid+1); 
	assert(node != NULL);
	int lid = node->b.lid_, leq = lid*bs_;
	if( (pb-pb0)%fisz != 0 ) jj = *pb++; // -2
	double *fpb = (double*)pb;
	for(jj=0 ; jj<bs_ ; jj++, leq++ ) ghostx[leq] = *fpb++;
	pb = (int*)fpb;
      }
#if defined(PROM_MEASURE_MPATH)
      int path = *pb++ + 1; 
      if( path > blck->path_ ) blck->path_ = path;
      if( (pb-pb0)%fisz != 0 ) jj = *pb++; // -2
      assert(path>0);
#endif
    }
    assert(pb-pb0 == sz);
  }
  if( buffer != NULL ) PetscFree(buffer);
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromPCGaussSeidel::SendISblocks"
/* PromPCGaussSeidel::SendISblocks ***********************************
 *
 *   INPUT:
 *    - proc_list
 *    - smessages
 *    - xx
 *    - tag
 *
 *   SIDE EFFECTS:
 *   RETURN:
 *     - PETSc error code
 */
int PromPCGaussSeidel::SendISblocks(PromTable<PromList<PromBlockXComm*>*>
				    *proc_list, 
				    PromList<MPI_Request*> *smessages,
				    const PromVector *const xx, 
				    const int tag ) const
{
  int       ierr,part,tt=sizeof(double),ii,jj,lid,eq,neq;
  ierr = MatGetOwnershipRange( A_->mat_, &ii, &jj ); CHKERRQ( ierr );
  const int myb0 = ii/bs_; jj = sizeof(MPI_Request); ii = sizeof(int);
  const int fisz = tt/ii, reqsz = (jj/tt + ((jj%tt == 0) ? 0 : 1)) * fisz;
  TablePos  tpos; double   *pv;    
  Position  pos;
  PromList<PromBlockXComm*> *list;
  
  ierr = xx->GetArray( &pv );        CHKERRQ(ierr);  
  tpos = proc_list->getHeadPosition(); 
  while( tpos ){ 
    list = proc_list->getNext(tpos, &part); part--;
    pos = list->getHeadPosition(); 
    if( pos == NULL ) continue; // none to send 
    tt = 0;
    while( pos != NULL ) {
      const PromBlockXComm * const commx = list->getNext(pos);
      tt += 2;
      for( ii = neq = 0 ; ii < commx->nNds_ ; ii++, neq += bs_ ){   // x only
	tt  += 1; // gid
	if( tt%fisz != 0 ) tt++;
	tt += bs_ * fisz; // x data
      }
#if defined(PROM_MEASURE_MPATH)
      tt += 1;
      if( tt%fisz != 0 ) tt++;
#endif
    }
    MPI_Request *sreq; ierr = PetscMalloc( (tt+reqsz)*sizeof(int),&sreq ); 
    CHKERRQ(ierr);
    smessages->addTail( sreq );
    int *buff = (int*)sreq + reqsz, *pb = buff;
    while( !list->isEmpty() ) {
      const PromBlockXComm * const commx = list->removeHead();
      *pb++ = commx->gbid_; *pb++ = commx->nNds_;
      for( ii = neq = 0 ; ii < commx->nNds_ ; ii++, neq += bs_ ){   // x only
	lid = commx->locBoundIDs_[ii]; 
	*pb++ = lid + myb0;
	eq = lid * bs_;
	if( (pb-buff)%fisz != 0 ) *pb++ = -2;
	double *fpb = (double*)pb;
	for( jj = 0 ; jj < bs_ ; jj++, eq++ ) *fpb++ = (double)pv[eq];
	pb = (int*)fpb;
      }
      assert(neq == commx->nEqs_);
      assert(tt >= pb-buff);
#if defined(PROM_MEASURE_MPATH)
      *pb++ = commx->path_; 
      assert(commx->path_>=0);
      if( (pb-buff)%fisz != 0 ) *pb++ = -2;
#endif
    }
    assert(tt == pb-buff);
    // send(tag) *< gbid, nn, nn*<gid, [-2,] ndf*<val> > >
    MPI_Isend( buff, tt, MPI_INT, part, tag, comm_, sreq );
  }
  ierr = xx->RestoreArray( &pv );     CHKERRQ(ierr);
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromPCGaussSeidel::PostRecvs"
/* PromPCGaussSeidel::PostRecvs ***********************************
 *
 *   INPUT:
 *    - commxList:
 *
 *   SIDE EFFECTS:
 *   RETURN:
 *     - PETSc error code
 */
int PromPCGaussSeidel::PostRecvs( const PromAList<PromProcXComm*> *commxList )
  const
{
  Position pos = commxList->getHeadPosition(); 
  while( pos ){
    PromProcXComm *cx = commxList->getNext(pos); 
    assert(cx->req_ == MPI_REQUEST_NULL);
    MPI_Irecv( cx->xarr_, cx->nEqs_ * sizeof(double)/sizeof(int), MPI_INT,
	       cx->part_, cx->tag_, cx->comm_, &cx->req_ );
  }
  return 0;
}
  
#undef __FUNCT__  
#define __FUNCT__ "PromPCGaussSeidel::SendLocalx"
/* PromPCGaussSeidel::SendLocalx ***********************************
 *
 *   INPUT:
 *    - commxList:
 *    - xx: x vector to read from
 *
 *   SIDE EFFECTS:
 *   RETURN:
 *     - PETSc error code
 */
int PromPCGaussSeidel::SendLocalx(const PromList<PromProcXComm*> *commxList,
				  const PromVector * const xx ) const
{
  assert(this!=NULL);
  int ierr,ii,jj,kk;
  double *pv;
  Position pos;
  
  ierr = xx->GetArray( &pv );        CHKERRQ(ierr);  
  pos = commxList->getHeadPosition(); 
  while( pos ){
    PromProcXComm *commx = commxList->getNext(pos);
    assert(commx->req_ == MPI_REQUEST_NULL);
    for( ii = kk = 0 ; ii < commx->nNds_ ; ii++ ) {
      int lbid = commx->lids_[ii], leq = lbid * bs_;
      for( jj = 0; jj < bs_; jj++, kk++, leq++ ) commx->xarr_[kk] = pv[leq];
    }
    assert(kk == commx->nEqs_); 
    // send(tag) < n*bs_*<val> >
    MPI_Isend(commx->xarr_, commx->nEqs_ * sizeof(double)/sizeof(int), 
	      MPI_INT, commx->part_, commx->tag_, commx->comm_,  &commx->req_);
  }
  ierr = xx->RestoreArray( &pv );     CHKERRQ(ierr);
  return 0;
} 

#undef __FUNCT__  
#define __FUNCT__ "PromPCGaussSeidel::RecvDoGhostx"
/* PromPCGaussSeidel::RecvDoGhostx ***********************************
 *
 *   - revieve updates (and process doable nodes emediately if bb != NULL)
 *
 *   INPUT:
 *    - commxList:
 *    - ghostx: ghost x values (out)
 *    - bb: (in/out)
 *    - xx: (in/out)
 *    - ndone: (in/out)
 *
 *   SIDE EFFECTS: 
 *   RETURN:
 *     - PETSc error code
 */
int PromPCGaussSeidel::RecvDoGhostx( PromAList<PromProcXComm*> *const commxList,
				     const PromVector *const bb /*=NULL*/, 
				     PromVector *const xx /*=NULL*/ ) const
{
  int        ii,jj,kk,lbid,ierr,gheq;
  const int  nghosts = gbid_ghost_.getCount();
  MPI_Status status;
  Position   pold, pos = NULL;

  bool exact_order = FALSE, reorder = FALSE;
  if( bb != NULL ) {
    if( commxList->count_++ == 0 ){
      bool det;
      ierr=options_.HasOptionName("-prometheus_use_deterministic",&det);
      if( det ) exact_order = TRUE;
      else reorder = TRUE; // semi stationary
    }
    else exact_order = TRUE;
  }

  // doit
  if( exact_order ) pos = commxList->getHeadPosition();  
  for( int xi = 0, todo = commxList->getCount() ; xi < todo ; xi++ ) {
    // get commx
    PromProcXComm *commx = NULL;
    if( exact_order ) {
      commx = commxList->getNext( pos );
      assert( commx->req_ != MPI_REQUEST_NULL );
      MPI_Wait( &commx->req_, &status );
    }
    else {
      while( commx == NULL ) {
	pos = commxList->getHeadPosition(); 
	while( (pold=pos) ){
	  PromProcXComm *cx = commxList->getNext( pos );
	  if( cx->req_ != MPI_REQUEST_NULL ) {
	    MPI_Test( &cx->req_, &ii, &status );
	    if( ii ) { 
	      assert( cx->req_ == MPI_REQUEST_NULL );
	      if( reorder ) {
		commxList->removeAt( pold );
		commxList->addTail( cx );
	      }
	      commx = cx; // outer 'break'
	      break;
	    }
	  }
	}
      }
      MPI_Wait( &commx->req_, &status );
    }
    // read in vals    
    for( ii = kk = 0 ; ii < commx->nNds_ ; ii++ ) {
      lbid = commx->lids_[ii];  assert(lbid<nghosts);
      gheq = lbid*bs_;
      for(jj=0 ; jj<bs_; jj++,kk++,gheq++) ghostx_[gheq] = commx->xarr_[kk];
    }
    // do it
    if( bb != NULL ) {
      ierr = DoBlocks( commx->InfBlocks_.arr_size_, 
		       commx->InfBlocks_.array_, 
		       bb, xx ); CHKERRQ(ierr);
    }
  }

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromPCGaussSeidel::WaitForList"
/* PromPCGaussSeidel::WaitForList ***********************************
 *
 *   INPUT:
 *    - commxList:
 *
 *   SIDE EFFECTS:
 *   RETURN:
 *     - PETSc error code
 */
int PromPCGaussSeidel::WaitForList(const PromList<PromProcXComm*> *commxList)
  const
{
  MPI_Status status; 
  
  Position pos = commxList->getHeadPosition(); 
  while( pos ){
    PromProcXComm *commx = commxList->getNext(pos);
    assert(commx->req_ != MPI_REQUEST_NULL);
    MPI_Wait( &commx->req_, &status );
  }
  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "PromPCGaussSeidel::GetProc" 
/* PromPCGaussSeidel::GetProc ************************************************
 *
 *  - find proc for global index
 *
 *  INPUT: 
 *    -  gid:
 *    -  proc: (out)
 *
 *  RETURN:
 *   - PETSc error code
 */
int PromPCGaussSeidel::GetProc( const int gbid, int &proc ) const
{
  const PromMap &map = A_->map_;
  const int np=map.comm_.np(),mype=map.comm_.mype(),fact=A_->isComplex() ? 2 : 1;
  int gid; const int*const p_gid = map.proc_gnode();
  if( !map.isConstNDF() ) { // convert bgid to gid, set 'nGlobalNd'
    const int myb0 = map.my0Eq()*fact/bs_,nextb0=myb0+map.nlocalEq()*fact/bs_;
    if( gbid < myb0 || gbid >= nextb0 ){
      const PromGSNode*const node = gbid_ghost_.find(gbid+1); 
      assert(node!=NULL);      gid = node->global(); 
    }
    else {
      proc = mype;
      return 0;
    }
  }
  else gid = gbid;
  //
  const int nGlobalNd = map.nglobal();
  if(gid<0||gid >= nGlobalNd)SETERRQ2(1,"gid<0||gid(%d)>=nGlobalNd(%d)",gid,nGlobalNd);
  float fproc = (float)gid*(float)np/(float)nGlobalNd + 0.5;
  proc = (int)fproc; assert(proc>=0);
  if( proc > np ) proc = np;
  while( gid < p_gid[proc] || gid >= p_gid[proc+1] ) {
    if( gid < p_gid[proc] ) proc--;
    else proc++;
  }
  if(proc<0 || proc>=np){ SETERRQ(1,"proc < 0 || proc >= numprocs"); }

  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "PromPCGaussSeidel::SetupNodalGS" 
/* PromPCGaussSeidel::SetupNodalGS *******************************************
 *
 *  INPUT: 
 *
 *  RETURN:
 *   - PETSc error code
 */
int PromPCGaussSeidel::SetupNodalGS()
{
  int  ierr,xx,ncols,jj,igid,adjacs[PROM_MAXNIEGH],done,flp;
  double *pv;  assert(bs_ > 0);
  double *Ai[PROM_MAXNIEGH]; assert(pc_ == NULL);
  const PromMatrix * const AA = A_; assert(AA!=NULL);
  ierr = MatGetOwnershipRange( A_->mat_, &xx, &jj); CHKERRQ( ierr );
  const int myb0=xx/bs_,nextb0=jj/bs_,nbloc=(nextb0-myb0);
  const int bs2=bs_*bs_;
  if(nGSBlocks_ != nbloc) SETERRQ2(1,"nGSBlocks_(%d) != nbloc(%d)",nGSBlocks_,nbloc); 
  if( nGSBlocks_ == 0 ) return 0;  
  
  if( NodalFactors_ == NULL ) {
    ierr = PetscMalloc((nGSBlocks_+1)*sizeof(int),&NodalFactorPointers_);
    CHKERRQ(ierr);
    for(xx=jj=igid=0, NodalFactorPointers_[0]=0 ; xx < nGSBlocks_ ; xx++ ){
      jj += bs2; igid += bs_;
      NodalFactorPointers_[xx+1] = jj;
    }
    ierr = PetscMalloc(jj*sizeof(double),&NodalFactors_);   CHKERRQ(ierr);
  } 

  for( xx = flp = 0, igid = myb0, pv = NodalFactors_ ; 
       xx < nGSBlocks_ ; xx++, igid++ ){
    ierr = AA->GetLocal_constBlk_NodeRow(xx, PROM_MAXNIEGH, ncols, adjacs,Ai);
    CHKERRQ(ierr);
    for( jj = 0, done = 0 ; jj < ncols ; jj++ ) {
      int jgid = adjacs[jj];
      if( igid == jgid ) {
	const double * const aijT = Ai[jj];
	ierr = PromContext::TInvert(aijT,pv,bs_); CHKERRQ(ierr);
	done++; flp += bs2;
	break;
      }
    }
    if(done==1) pv += bs2;
    else {
      assert(done==0);
      for(int i=0;i<bs_;i++)for(int j=0;j<bs_;j++) *pv++ = (i==j) ? 1. : 0.;
    }
    flp += bs_*bs2;
  }
  perf_mon_.Flops( 2*flp );

  return 0;
}

#if defined(PROM_USE_PETSC)

#undef __FUNCT__  
#define __FUNCT__ "PromPCGaussSeidel::DoBlocks"
/* PromPCGaussSeidel::DoBlocks ***********************************
 *
 *   - apply blocks of GS smoothing 
 *
 *   INPUT:
 *    - nb
 *    - blocks[nb]: list of blocks (can be empty)
 *    - bb: 
 *    - xx; 
 *
 *   SIDE EFFECTS:
 * 
 *   RETURN:
 *     - PETSc error code
 */
int PromPCGaussSeidel::DoBlocks( const int nb, const int blocks[],
				 const PromVector * const bb,
				 PromVector *const xx ) const
{
  if( nb == 0 ) return 0;  
  if( A_->isSymmetricMatrix() )SETERRQ(1,"Symmetric matrix not implemented");

  if( pc_ != NULL ) {
    switch( bs_ ) {
    case 3: return DoBlocks_schwarz_3(nb, blocks, bb, xx, pc_);
    case 6: return DoBlocks_schwarz_6(nb, blocks, bb, xx, pc_);
    case 1: return DoBlocks_schwarz_1(nb, blocks, bb, xx, pc_);
    default : 
      return DoBlocks_schwarz(nb, blocks, bb, xx, pc_, bs_ );
    }
  }
  else {
    //
    switch( bs_ ) {
    case 3: return DoBlocks_nodal_3(nb,blocks,bb,xx);
    case 6: return DoBlocks_nodal_6(nb,blocks,bb,xx);
    case 2: return DoBlocks_nodal_2(nb,blocks,bb,xx);
    case 1: return DoBlocks_nodal_1(nb,blocks,bb,xx);
    default : 
      return DoBlocks_nodal(nb,blocks,bb,xx,bs_);
    }
  }
}

#undef __FUNCT__  
#define __FUNCT__ "PromPCGaussSeidel::DoBlocks_nodal"
/* PromPCGaussSeidel::DoBlocks_nodal ***********************************
 *
 *   - apply blocks of GS smoothing (PETSc)
 *
 *   INPUT:
 *    - nb
 *    - blocks[nb]: list of blocks (can be empty)
 *    - bb: 
 *    - xx; 
 *
 *   SIDE EFFECTS:
 * 
 *   RETURN:
 *     - PETSc error code
 */
int PromPCGaussSeidel::DoBlocks_nodal( const int nb, const int blocks[],
				       const PromVector *const bb, 
				       PromVector *const xx, 
				       const int bs ) const 
{  
  int       ierr,ii,jj,colj;
  const int pre_s=pre_smooth_,bs2=bs*bs;
  int       numops,kinc=(pre_s ? 1 : -1);
  double    temp, *barr, *xarr,mAx[2*PROM_MAXNDF];
  const double *px; 
  const bool havgh = (A_->map_.comm_.np() > 1);
  Mat AA = A_->mat_; 
  const double *aa=NULL, *gh_aa = NULL;
  const int *ai,*aj,*ailen,*gh_ai=NULL,*gh_aj=NULL,*gh_ailen=NULL,*garr=NULL;

  perf_mon_.EventBegin(PromContext::prom_logs_[LOG_FINALIZE]);
  if( bs == 1 ) {
    Mat_SeqAIJ *mat;
    if( havgh ) {
      Mat_MPIAIJ *aij = (Mat_MPIAIJ*)AA->data; // upper left matix
      garr = aij->garray;
      mat = (Mat_SeqAIJ*)aij->B->data;
      // get ghost pointers
      gh_ai = mat->i; gh_aj = mat->j; gh_ailen = mat->ilen;   gh_aa = mat->a;
      //
      mat = (Mat_SeqAIJ*)aij->A->data; 
    }
    else {
      mat = (Mat_SeqAIJ*)AA->data;
    }
    // get local pointers
    ai = mat->i; aj = mat->j; ailen = mat->ilen;   aa = mat->a;
  }
  else {
    Mat_SeqBAIJ *mat;
    if( havgh ) {
      Mat_MPIBAIJ *baij = (Mat_MPIBAIJ*)AA->data; // upper left matix
      garr = baij->garray;
      mat = (Mat_SeqBAIJ*)baij->B->data;
      // get ghost pointers
      gh_ai = mat->i; gh_aj = mat->j; gh_ailen = mat->ilen;   gh_aa = mat->a;
      //
      mat = (Mat_SeqBAIJ*)baij->A->data; 
    }
    else {
      mat = (Mat_SeqBAIJ*)AA->data;
    }
    // get local pointers
    ai = mat->i; aj = mat->j; ailen = mat->ilen;   aa = mat->a;
  }
  
  ierr = bb->GetArray( &barr );        CHKERRQ(ierr);
  ierr = xx->GetArray( &xarr );       CHKERRQ(ierr); 
  for(int kkk = numops = 0, kk = (pre_s ? 0 : nb-1); kkk<nb; kkk++,kk += kinc){
    const int lidi = blocks[kk];
    // y = b_i
    for(ii = 0, jj = bs*lidi ; ii < bs ; ii++, jj++ ) mAx[ii] = barr[jj]; 
    int pp = ai[lidi], ncols = ailen[lidi]; const int *pj = aj + pp;
    const double * pa = aa + bs2*pp;
    // y = y - Ax (local,ghost)
    numops += ncols;
    for( colj = 0 ; colj != ncols ; colj++ ) {
      int lidj = *pj++;
      if( lidj == lidi ){ pa += bs2; continue; }
      else px = &xarr[lidj*bs]; // local 
      // b_i -= A_ij * x_j
      for( ii = 0 ; ii < bs ; ii++ ) {
	temp = px[ii];
	for( jj = 0 ; jj < bs ; jj++ ) mAx[jj] -= (*pa++) * temp;
      }
    }
    // inc ghosts
    if( havgh ) {
      pp = gh_ai[lidi]; pj = gh_aj + pp; ncols = gh_ailen[lidi];
      pa = gh_aa + bs2*pp;
      // y = y - Ax (local,ghost)
      numops += ncols;
      for( colj = 0 ; colj != ncols ; colj++ ) {
	int lidj = *pj++;
	int gid = garr[lidj]; // PETSc's idea of global IDs
	PromGSNode *node = gbid_ghost_.find(gid+1); 
	if(node == NULL) { pa += bs2; continue; }
	lidj = node->b.lid_;
	px = &ghostx_[lidj*bs]; 
	// b_i -= A_ij * x_j
	for( ii = 0 ; ii < bs ; ii++ ) {
	  temp = px[ii];
	  for( jj = 0 ; jj < bs ; jj++ ) mAx[jj] -= (*pa++) * temp;
	}
      }
    }
    // do Aii^-1
    const double *pv = NodalFactors_ + NodalFactorPointers_[lidi]; 
    for( ii = 0, jj = bs*lidi ; ii < bs ; ii++, jj++ ) {
      temp = 0.0; 
      for(int kk = 0 ; kk < bs ; kk++ ) temp += *pv++ *  mAx[kk];
      xarr[jj] = temp;
    }
  }
  ierr = xx->RestoreArray( &xarr );     CHKERRQ(ierr); 
  ierr = bb->RestoreArray( &barr );     CHKERRQ(ierr); 
  //
  perf_mon_.Flops(2*bs2*numops);
  perf_mon_.EventEnd(PromContext::prom_logs_[LOG_FINALIZE]);
   
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromPCGaussSeidel::DoBlocks_nodal_1"
int PromPCGaussSeidel::DoBlocks_nodal_1( const int nb, const int blocks[],
					 const PromVector *const bb, 
					 PromVector *const xx ) const
{  
  int       ierr,colj;
  const int pre_s=pre_smooth_;
  int       numops,kinc=(pre_s ? 1 : -1);
  double   *barr, *xarr, mAx;
  const double *px;
  const bool havgh = (A_->map_.comm_.np() > 1);
  Mat AA = A_->mat_; 
  const double *aa=NULL, *gh_aa = NULL;
  const int *ai,*aj,*ailen,*gh_ai=NULL,*gh_aj=NULL,*gh_ailen=NULL,*garr=NULL;
  Mat_SeqAIJ  *mat;

  perf_mon_.EventBegin(PromContext::prom_logs_[LOG_FINALIZE]);
  // get matrix data
  if( havgh ) {
    Mat_MPIAIJ *aij =(Mat_MPIAIJ*)AA->data;
    garr = aij->garray;
    mat = (Mat_SeqAIJ*)aij->B->data; 
    // get ghost pointers
    gh_ai = mat->i; gh_aj = mat->j; gh_ailen = mat->ilen; gh_aa = mat->a;
    //
    mat = (Mat_SeqAIJ*)aij->A->data; 
  }
  else {
    mat = (Mat_SeqAIJ*)AA->data; 
  }
  ai = mat->i; aj = mat->j; ailen = mat->ilen;  aa = mat->a;
  
  ierr = bb->GetArray( &barr );        CHKERRQ(ierr);
  ierr = xx->GetArray( &xarr );       CHKERRQ(ierr); 
  for(int kkk = numops = 0, kk = (pre_s ? 0 : nb-1); kkk<nb; kkk++,kk += kinc){
    const int lidi = blocks[kk];
    // y = b_i 
    mAx = barr[lidi]; 
    int pp = ai[lidi], ncols = ailen[lidi]; const int *pj = aj + pp;
    const double * pa = aa + pp;
    // y = y - Ax (local,ghost)
    numops += ncols;
    for( colj = 0 ; colj != ncols ; colj++ ) {
      int lidj = *pj++;
      if( lidj == lidi ){ pa += 1; continue; }
      else px = &xarr[lidj]; // local 
      // b_i -= A_ij * x_j
      mAx -= (*pa++) * (*px);
    }
    if( havgh ) {
      pp = gh_ai[lidi]; pj = gh_aj + pp; ncols = gh_ailen[lidi];
      pa = gh_aa + pp;
      // y = y - Ax (ghost)
      numops += ncols;
      for( colj = 0 ; colj != ncols ; colj++ ) {
	int lidj = *pj++;
	int gid = garr[lidj]; // PETSc's idea of global IDs
	PromGSNode *node = gbid_ghost_.find(gid+1); 
	if(node == NULL) { pa += 1; continue; }
	lidj = node->b.lid_;
	px = &ghostx_[lidj]; 
	// b_i -= A_ij * x_j
	mAx -= (*pa++) * (*px);
      }
    }
    // do Aii^-1
    const double *pv = NodalFactors_ + NodalFactorPointers_[lidi];
    //xarr[lidi] = one_omega * xarr[lidi] + *pv * mAx;
    xarr[lidi] = *pv * mAx;
  } 
  ierr = xx->RestoreArray( &xarr );     CHKERRQ(ierr); 
  ierr = bb->RestoreArray( &barr );     CHKERRQ(ierr); 
  //
  perf_mon_.Flops(2*numops);
  perf_mon_.EventEnd(PromContext::prom_logs_[LOG_FINALIZE]);
   
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromPCGaussSeidel::DoBlocks_nodal_2"
int PromPCGaussSeidel::DoBlocks_nodal_2( const int nb, const int blocks[],
					 const PromVector *const bb, 
					 PromVector *const xx ) const
{  
  int       ierr,ii,jj,colj;
  const int pre_s=pre_smooth_;
  int       numops,kinc=(pre_s ? 1 : -1);
  double    tt,p0,p1,*barr,*xarr,mAx[2];
  const double *px; 
  const bool havgh = (A_->map_.comm_.np() > 1);
  Mat AA = A_->mat_; 
  const double *aa=NULL, *gh_aa = NULL;
  const int *ai,*aj,*ailen,*gh_ai=NULL,*gh_aj=NULL,*gh_ailen=NULL,*garr=NULL;
  Mat_SeqBAIJ *mat;

  // get matrix data
  perf_mon_.EventBegin(PromContext::prom_logs_[LOG_FINALIZE]);
  if( havgh ) {
    Mat_MPIBAIJ *baij = (Mat_MPIBAIJ*)AA->data; // upper left matix
    garr = baij->garray;
    mat = (Mat_SeqBAIJ*)baij->B->data;
    // get ghost pointers
    gh_ai = mat->i; gh_aj = mat->j; gh_ailen = mat->ilen;   gh_aa = mat->a;
    //
    mat = (Mat_SeqBAIJ*)baij->A->data; 
  }
  else {
    mat = (Mat_SeqBAIJ*)AA->data;
  }
  ai = mat->i; aj = mat->j; ailen = mat->ilen;   aa = mat->a;

  ierr = bb->GetArray( &barr );        CHKERRQ(ierr);
  ierr = xx->GetArray( &xarr );       CHKERRQ(ierr); 
  for(int kkk = numops=0, kk = (pre_s ? 0 : nb-1); kkk<nb; kkk++,kk+=kinc){
    const int lidi = blocks[kk];
    // y = b_i
    for( ii=0, jj = 2*lidi ; ii < 2 ; ii++, jj++ ) mAx[ii] = barr[jj]; 
    int pp = ai[lidi], ncols = ailen[lidi]; const int *pj = aj + pp;
    const double * pa = aa + 4*pp;
    // y = y - Ax (local)
    numops += ncols;
    for( colj = 0 ; colj != ncols ; colj++ ) {
      int lidj = *pj++;
      if( lidj == lidi ){ pa += 4; continue; }
      else px = &xarr[lidj*2]; // local 
      // b_i -= A_ij * x_j
      tt=px[0]; p0 = (*pa++)*tt; p1 = (*pa++)*tt; 
      tt=px[1]; p0 +=(*pa++)*tt; p1 +=(*pa++)*tt; 
      mAx[0] -= p0; mAx[1] -= p1;  
    }
    // y = y - Ax (ghost)
    if( havgh ) {
      pp = gh_ai[lidi]; pj = gh_aj + pp; ncols = gh_ailen[lidi];
      pa = gh_aa + 4*pp;
      numops += ncols;
      for( colj = 0 ; colj != ncols ; colj++ ) {
	int lidj = *pj++;
	int gid = garr[lidj]; // PETSc's idea of global IDs
	PromGSNode *node = gbid_ghost_.find(gid+1); 
	if(node == NULL) { pa += 4; continue; }
	lidj = node->b.lid_;
	px = &ghostx_[lidj*2];
	// b_i -= A_ij * x_j
	tt=px[0]; p0 = (*pa++)*tt; p1 = (*pa++)*tt;
	tt=px[1]; p0 +=(*pa++)*tt; p1 +=(*pa++)*tt;
	mAx[0] -= p0; mAx[1] -= p1;
      }
    }
    // do Aii^-1
    const double *pv = NodalFactors_ + NodalFactorPointers_[lidi];
    tt = mAx[0]; p0 =  pv[0]*tt; p1  = pv[2]*tt;
    tt = mAx[1]; p0 += pv[1]*tt; p1 += pv[3]*tt;
    jj = 2*lidi;
    xarr[jj] = p0; 
    xarr[jj+1] = p1; 
  } 
  ierr = xx->RestoreArray( &xarr );     CHKERRQ(ierr); 
  ierr = bb->RestoreArray( &barr );     CHKERRQ(ierr); 
  //
  perf_mon_.Flops(2*4*numops);
  perf_mon_.EventEnd(PromContext::prom_logs_[LOG_FINALIZE]);
   
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromPCGaussSeidel::DoBlocks_nodal_3"
int PromPCGaussSeidel::DoBlocks_nodal_3( const int nb, const int blocks[],
					 const PromVector *const bb, 
					 PromVector *const xx ) const
{  
  int       ierr,ii,jj,colj;
  const int pre_s=pre_smooth_;
  int       numops,kinc=(pre_s ? 1 : -1);
  double    tt,p0,p1,p2,*barr,*xarr,mAx[3];
  const double *px; 
  const bool havgh = (A_->map_.comm_.np() > 1);
  Mat AA = A_->mat_; 
  const double *aa=NULL, *gh_aa = NULL;
  const int *ai,*aj,*ailen,*gh_ai=NULL,*gh_aj=NULL,*gh_ailen=NULL,*garr=NULL;
  Mat_SeqBAIJ *mat;

  // get matrix data
  perf_mon_.EventBegin(PromContext::prom_logs_[LOG_FINALIZE]);
  if( havgh ) {
    Mat_MPIBAIJ *baij = (Mat_MPIBAIJ*)AA->data; // upper left matix
    garr = baij->garray;
    mat = (Mat_SeqBAIJ*)baij->B->data;
    // get ghost pointers
    gh_ai = mat->i; gh_aj = mat->j; gh_ailen = mat->ilen;   gh_aa = mat->a;
    //
    mat = (Mat_SeqBAIJ*)baij->A->data; 
  }
  else {
    mat = (Mat_SeqBAIJ*)AA->data;
  }
  ai = mat->i; aj = mat->j; ailen = mat->ilen;   aa = mat->a;

  ierr = bb->GetArray( &barr );        CHKERRQ(ierr);
  ierr = xx->GetArray( &xarr );       CHKERRQ(ierr); 
  for(int kkk = numops = 0, kk = (pre_s ? 0 : nb-1); kkk<nb; kkk++,kk += kinc){
    const int lidi = blocks[kk];
    // y = b_i
    for( ii=0, jj = 3*lidi ; ii < 3 ; ii++, jj++ ) mAx[ii] = barr[jj]; 
    int pp = ai[lidi], ncols = ailen[lidi]; const int *pj = aj + pp;
    const double * pa = aa + 9*pp;
    // y = y - Ax (local)
    numops += ncols;
    for( colj = 0 ; colj != ncols ; colj++ ) {
      int lidj = *pj++;
      if( lidj == lidi ){ pa += 9; continue; }
      else px = &xarr[lidj*3]; // local 
      // b_i -= A_ij * x_j
      tt=px[0]; p0 = (*pa++)*tt; p1 = (*pa++)*tt; p2  = (*pa++)*tt;
      tt=px[1]; p0 +=(*pa++)*tt; p1 +=(*pa++)*tt; p2 += (*pa++)*tt;
      tt=px[2]; p0 +=(*pa++)*tt; p1 +=(*pa++)*tt; p2 += (*pa++)*tt;
      mAx[0] -= p0; mAx[1] -= p1;  mAx[2] -= p2; 
    }
    // y = y - Ax (ghost)
    if( havgh ) {
      pp = gh_ai[lidi]; pj = gh_aj + pp; ncols = gh_ailen[lidi];
      pa = gh_aa + 9*pp;
      numops += ncols;
      for( colj = 0 ; colj != ncols ; colj++ ) {
	int lidj = *pj++;
	int gid = garr[lidj]; // PETSc's idea of global IDs
	PromGSNode *node = gbid_ghost_.find(gid+1); 
	if(node == NULL) { pa += 9; continue; }
	lidj = node->b.lid_;
	px = &ghostx_[lidj*3];
	// b_i -= A_ij * x_j
	tt=px[0]; p0 = (*pa++)*tt; p1 = (*pa++)*tt; p2  = (*pa++)*tt;
	tt=px[1]; p0 +=(*pa++)*tt; p1 +=(*pa++)*tt; p2 += (*pa++)*tt;
	tt=px[2]; p0 +=(*pa++)*tt; p1 +=(*pa++)*tt; p2 += (*pa++)*tt;
	mAx[0] -= p0; mAx[1] -= p1;  mAx[2] -= p2; 
      }
    }
    // do Aii^-1
    const double *pv = NodalFactors_ + NodalFactorPointers_[lidi];
    tt = mAx[0]; p0 =  pv[0]*tt; p1  = pv[3]*tt; p2  = pv[6]*tt;
    tt = mAx[1]; p0 += pv[1]*tt; p1 += pv[4]*tt; p2 += pv[7]*tt;
    tt = mAx[2]; p0 += pv[2]*tt; p1 += pv[5]*tt; p2 += pv[8]*tt;
    jj = 3*lidi;
    //xarr[jj] = one_omega * xarr[jj] + p0; 
    //xarr[jj+1] = one_omega * xarr[jj+1] + p1; 
    //xarr[jj+2] = one_omega * xarr[jj+2] + p2; 
    xarr[jj] = p0; 
    xarr[jj+1] = p1; 
    xarr[jj+2] = p2; 
  } 
  ierr = xx->RestoreArray( &xarr );     CHKERRQ(ierr); 
  ierr = bb->RestoreArray( &barr );     CHKERRQ(ierr); 
  //
  perf_mon_.Flops(2*9*numops);
  perf_mon_.EventEnd(PromContext::prom_logs_[LOG_FINALIZE]);
   
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromPCGaussSeidel::DoBlocks_nodal_6"
int PromPCGaussSeidel::DoBlocks_nodal_6( const int nb, const int blocks[],
					 const PromVector *const bb, 
					 PromVector *const xx ) const
{  
  int       ierr,ii,jj,colj;
  const int pre_s=pre_smooth_;
  int       numops,kinc=(pre_s ? 1 : -1);
  double    tt,p0,p1,p2,p3,p4,p5,*barr,*xarr,mAx[6];
  const double *px; 
  const bool havgh = (A_->map_.comm_.np() > 1);
  Mat AA = A_->mat_; 
  const double *aa=NULL, *gh_aa = NULL;
  const int *ai,*aj,*ailen,*gh_ai=NULL,*gh_aj=NULL,*gh_ailen=NULL,*garr=NULL;
  Mat_SeqBAIJ *mat;

  perf_mon_.EventBegin(PromContext::prom_logs_[LOG_FINALIZE]);
  if( havgh ) {
    Mat_MPIBAIJ *baij = (Mat_MPIBAIJ*)AA->data; // upper left matix
    garr = baij->garray;
    mat = (Mat_SeqBAIJ*)baij->B->data;
    // get ghost pointers
    gh_ai = mat->i; gh_aj = mat->j; gh_ailen = mat->ilen;   gh_aa = mat->a;
    //
    mat = (Mat_SeqBAIJ*)baij->A->data; 
  }
  else {
    mat = (Mat_SeqBAIJ*)AA->data;
  }
  // get local pointers
  ai = mat->i; aj = mat->j; ailen = mat->ilen;   aa = mat->a;

  ierr = bb->GetArray( &barr );        CHKERRQ(ierr);
  ierr = xx->GetArray( &xarr );       CHKERRQ(ierr); 
  for(int kkk = numops = 0, kk = (pre_s ? 0 : nb-1); kkk<nb; kkk++,kk += kinc){
    const int lidi = blocks[kk];
    // y = b_i
    for( ii=0, jj = 6*lidi ; ii < 6 ; ii++, jj++ ) mAx[ii] = barr[jj]; 
    int pp = ai[lidi], ncols = ailen[lidi]; const int *pj = aj + pp;
    const double * pa = aa + 36*pp;
    // y = y - Ax (local)
    numops += ncols;
    for( colj = 0 ; colj != ncols ; colj++ ) {
      int lidj = *pj++;
      if( lidj == lidi ){ pa += 36; continue; }
      else px = &xarr[lidj*6]; // local 
      // b_i -= A_ij * x_j
      tt = px[0]; 
      p0 = (*pa++)*tt; p1 = (*pa++)*tt; p2 = (*pa++)*tt;
      p3 = (*pa++)*tt; p4 = (*pa++)*tt; p5 = (*pa++)*tt;
      tt = px[1]; 
      p0 += (*pa++)*tt; p1 += (*pa++)*tt; p2 += (*pa++)*tt;
      p3 += (*pa++)*tt; p4 += (*pa++)*tt; p5 += (*pa++)*tt;
      tt = px[2]; 
      p0 += (*pa++)*tt;p1 += (*pa++)*tt;p2 += (*pa++)*tt;
      p3 += (*pa++)*tt;p4 += (*pa++)*tt;p5 += (*pa++)*tt;
      tt = px[3]; 
      p0 += (*pa++)*tt;p1 += (*pa++)*tt;p2 += (*pa++)*tt;
      p3 += (*pa++)*tt;p4 += (*pa++)*tt;p5 += (*pa++)*tt;
      tt = px[4]; 
      p0 += (*pa++)*tt;p1 += (*pa++)*tt;p2 += (*pa++)*tt;
      p3 += (*pa++)*tt;p4 += (*pa++)*tt;p5 += (*pa++)*tt;
      tt = px[5]; 
      p0 += (*pa++)*tt;p1 += (*pa++)*tt;p2 += (*pa++)*tt;
      p3 += (*pa++)*tt;p4 += (*pa++)*tt;p5 += (*pa++)*tt;
      mAx[0] -= p0; mAx[1] -= p1; mAx[2] -= p2; 
      mAx[3] -= p3; mAx[4] -= p4; mAx[5] -= p5;
    }
    // y = y - Ax (ghost)
    if( havgh ) {
      pp = gh_ai[lidi]; pj = gh_aj + pp; ncols = gh_ailen[lidi];
      pa = gh_aa + 36*pp;
      numops += ncols;
      for( colj = 0 ; colj != ncols ; colj++ ) {
	int lidj = *pj++;
	int gid = garr[lidj]; // PETSc's idea of global IDs
	PromGSNode *node = gbid_ghost_.find(gid+1); 
	if(node == NULL) { pa += 36; continue; }
	lidj = node->b.lid_;
	px = &ghostx_[lidj*6]; 
	// b_i -= A_ij * x_j
	tt = px[0]; 
	p0 = (*pa++)*tt; p1 = (*pa++)*tt; p2 = (*pa++)*tt;
	p3 = (*pa++)*tt; p4 = (*pa++)*tt; p5 = (*pa++)*tt;
	tt = px[1]; 
	p0 += (*pa++)*tt; p1 += (*pa++)*tt; p2 += (*pa++)*tt;
	p3 += (*pa++)*tt; p4 += (*pa++)*tt; p5 += (*pa++)*tt;
	tt = px[2]; 
	p0 += (*pa++)*tt;p1 += (*pa++)*tt;p2 += (*pa++)*tt;
	p3 += (*pa++)*tt;p4 += (*pa++)*tt;p5 += (*pa++)*tt;
	tt = px[3]; 
	p0 += (*pa++)*tt;p1 += (*pa++)*tt;p2 += (*pa++)*tt;
	p3 += (*pa++)*tt;p4 += (*pa++)*tt;p5 += (*pa++)*tt;
	tt = px[4]; 
	p0 += (*pa++)*tt;p1 += (*pa++)*tt;p2 += (*pa++)*tt;
	p3 += (*pa++)*tt;p4 += (*pa++)*tt;p5 += (*pa++)*tt;
	tt = px[5]; 
	p0 += (*pa++)*tt;p1 += (*pa++)*tt;p2 += (*pa++)*tt;
	p3 += (*pa++)*tt;p4 += (*pa++)*tt;p5 += (*pa++)*tt;
	mAx[0] -= p0; mAx[1] -= p1; mAx[2] -= p2; 
	mAx[3] -= p3; mAx[4] -= p4; mAx[5] -= p5;
      }
    }
    // do Aii^-1
    const double *pv = NodalFactors_ + NodalFactorPointers_[lidi];
    tt = mAx[0]; 
    p0 = pv[0]*tt;   p1 = pv[6]*tt;   p2 = pv[12]*tt;
    p3 = pv[18]*tt;  p4 = pv[24]*tt;  p5 = pv[30]*tt;
    tt = mAx[1];
    p0 += pv[1]*tt;  p1 += pv[7]*tt;  p2 += pv[13]*tt;
    p3 += pv[19]*tt; p4 += pv[25]*tt; p5 += pv[31]*tt;
    tt = mAx[2]; 
    p0 += pv[2]*tt;  p1 += pv[8]*tt;  p2 += pv[14]*tt;
    p3 += pv[20]*tt; p4 += pv[26]*tt; p5 += pv[32]*tt;
    tt = mAx[3]; 
    p0 += pv[3]*tt;  p1 += pv[9]*tt;  p2 += pv[15]*tt;
    p3 += pv[21]*tt; p4 += pv[27]*tt; p5 += pv[33]*tt;
    tt = mAx[4]; 
    p0 += pv[4]*tt;  p1 += pv[10]*tt; p2 += pv[16]*tt;
    p3 += pv[22]*tt; p4 += pv[28]*tt; p5 += pv[34]*tt;
    tt = mAx[5]; 
    p0 += pv[5]*tt;  p1 += pv[11]*tt; p2 += pv[17]*tt;
    p3 += pv[23]*tt; p4 += pv[29]*tt; p5 += pv[35]*tt;
    jj = 6*lidi;
    /*xarr[jj]   = one_omega * xarr[jj] + p0; 
      xarr[jj+1] = one_omega * xarr[jj+1] + p1; 
      xarr[jj+2] = one_omega * xarr[jj+2] + p2; 
      xarr[jj+3] = one_omega * xarr[jj+3] + p3; 
      xarr[jj+4] = one_omega * xarr[jj+4] + p4; 
      xarr[jj+5] = one_omega * xarr[jj+5] + p5;	*/
    xarr[jj]   = p0; 
    xarr[jj+1] = p1; 
    xarr[jj+2] = p2; 
    xarr[jj+3] = p3; 
    xarr[jj+4] = p4; 
    xarr[jj+5] = p5;
  }
  ierr = xx->RestoreArray( &xarr );     CHKERRQ(ierr); 
  ierr = bb->RestoreArray( &barr );     CHKERRQ(ierr); 
  //
  perf_mon_.Flops(2*36*numops);
  perf_mon_.EventEnd(PromContext::prom_logs_[LOG_FINALIZE]);
   
  return 0;  
}

#define PROM_MAX_BLK_SZ 10000
#undef __FUNCT__  
#define __FUNCT__ "PromPCGaussSeidel::DoBlocks_schwarz"
/* PromPCGaussSeidel::DoBlock_schwarz ***********************************
 *
 *   - apply blocks of GS smoothing (PETSc)
 *
 *   INPUT:
 *    - nb
 *    - blocks[nb]: list of blocks (can be empty)
 *    - bb: 
 *    - xx; 
 *    - pc:
 *    - bs:
 *
 *   SIDE EFFECTS:
 * 
 *   RETURN:
 *     - PETSc error code
 */
int PromPCGaussSeidel::DoBlocks_schwarz( const int nb, const int blocks[],
					 const PromVector *const bb, 
					 PromVector *const xx,
					 const PromPC * const pc, 
					 const int bs ) const
{
  if( nb == 0 ) return 0;
  PETSc_ASM *osm = (PETSc_ASM*)pc_->pc_->data;             assert(osm != NULL);
  Mat AA = A_->mat_;  int ierr, numops,myb0eq,nextb0eq;
  ierr = MatGetOwnershipRange( A_->mat_, &myb0eq, &nextb0eq); CHKERRQ( ierr );
  const bool havgh = (A_->map_.comm_.np() > 1);
  perf_mon_.EventBegin(PromContext::prom_logs_[LOG_FINALIZE]);
  
  ierr = PromDoSchwarzBlocks(AA, nb, blocks, osm, bb, xx, bs, myb0eq/bs_, myb0eq, 
			     havgh, pre_smooth_, &gbid_ghost_, ghostx_, numops);
  CHKERRQ(ierr);
  //
  perf_mon_.Flops(2*bs*bs*numops);
  perf_mon_.EventEnd(PromContext::prom_logs_[LOG_FINALIZE]);
 
  return 0;
}
  
#undef __FUNCT__  
#define __FUNCT__ "PromDoSchwarzBlocks"
/* PromDoSchwarzBlocks ***********************************
 *
 *   - apply blocks of GS smoothing (PETSc)
 *
 *   INPUT:
 *    - nb
 *    - blocks[nb]: list of blocks (can be empty)
 *    - bb: 
 *    - xx; 
 *    - bs:
 *    - myb0:
 *    - my0Eq:
 *    - havgh:
 *    - pre_s:
 *    - gbid_ghost:
 *    - ghostx:
 *    - numops: (out)
 *
 *   SIDE EFFECTS:
 * 
 *   RETURN:
 *     - PETSc error code
 */
int PromDoSchwarzBlocks( Mat AA, const int nb,const int blocks[],
			 PETSc_ASM *osm, const PromVector *const bb,
			 PromVector *const xx, const int bs, const int myb0,
			 const int my0Eq, const bool havgh, const bool pre_s, 
			 const PromTable<PromGSNode*> *gbid_ghost, 
			 const double ghostx[], int &numops ) 
{
  int          ierr,ii,jj,colj,ncols,nnodes,eq,pp,nneq,nn;
  const int    *geqs;
  const double*aa=NULL,*gh_aa=NULL,*pa,*px;
  const int   *ai,*aj,*ailen,*gh_ai=NULL,*gh_aj=NULL,*gh_ailen=NULL,*garr=NULL;
  const int    bs2=bs*bs,*pj,kinc=(pre_s ? 1:-1),ksrt=(pre_s ? 0:nb-1);  
  double      *barr,*xarr,*yarr,*pv,tt,tarr[2*PROM_MAXNDF];
  int          gids[PROM_MAX_BLK_SZ];
  
  if( bs == 1 ) {
    Mat_SeqAIJ *mat;
    if( havgh ) {
      Mat_MPIAIJ *aij = (Mat_MPIAIJ*)AA->data; // upper left matix
      garr = aij->garray;
      mat = (Mat_SeqAIJ*)aij->B->data;
      // get ghost pointers
      gh_ai = mat->i; gh_aj = mat->j; gh_ailen = mat->ilen;   gh_aa = mat->a;
      //
      mat = (Mat_SeqAIJ*)aij->A->data; 
    }
    else {
      mat = (Mat_SeqAIJ*)AA->data;
    }
    // get local pointers
    ai = mat->i; aj = mat->j; ailen = mat->ilen;   aa = mat->a;
  }
  else {
    Mat_SeqBAIJ *mat;
    if( havgh ) {
      Mat_MPIBAIJ *baij = (Mat_MPIBAIJ*)AA->data; // upper left matix
      garr = baij->garray;
      mat = (Mat_SeqBAIJ*)baij->B->data;
      // get ghost pointers
      gh_ai = mat->i; gh_aj = mat->j; gh_ailen = mat->ilen;   gh_aa = mat->a;
      //
      mat = (Mat_SeqBAIJ*)baij->A->data; 
    }
    else {
      mat = (Mat_SeqBAIJ*)AA->data;
    }
    // get local pointers
    ai = mat->i; aj = mat->j; ailen = mat->ilen;   aa = mat->a;
  }

  ierr = bb->GetArray( &barr );       CHKERRQ(ierr);
  ierr = xx->GetArray( &xarr );       CHKERRQ(ierr);
  numops = 0;
  for( int xxx = 0, kk = ksrt ; xxx < nb ; xxx++, kk += kinc) {
    const int bidi = blocks[kk];
    // get rows of block
    IS is = osm->is[bidi];
    KSP ksp = osm->ksp[bidi];
    Vec x = osm->x[bidi]; Vec y = osm->y[bidi];
    ierr = VecGetArray( y, &yarr );       CHKERRQ(ierr);
    ierr = ISGetSize( is, &nneq ); CHKERRQ(ierr); nnodes = nneq/bs;
    if(nnodes>PROM_MAX_BLK_SZ)SETERRQ2(1,"PROM_MAX_BLK_SZ(=%d) < bsz=%d",
				       PROM_MAX_BLK_SZ,nnodes);
    ierr = ISGetIndices( is, &geqs );                       CHKERRQ(ierr);
    // get gids (mask)
    for( nn = eq = 0 ; nn < nnodes ; nn++, eq += bs) gids[nn] = geqs[eq]/bs;
    // y <- b - Ax
    for( nn = eq = 0 ; nn < nnodes ; nn++, eq += bs ) {
      const int geq = geqs[eq], gidi = geq/bs, lidi = gidi - myb0; 
      // get part (A) and 'bb' into temp array 'tarr'  
      pp = ai[lidi]; pj = aj + pp; // part (A)
      pa = aa + bs2*pp;  ncols = ailen[lidi]; 
      px = &barr[lidi*bs]; 
      // tarr <- b
      for(ii=0 ; ii<bs ; ii++) tarr[ii] = px[ii]; 
      // y = y - Ax (local)
      for( colj = 0 ; colj != ncols ; colj++ ) {
	int lidj = *pj++;
	int doit=1,gidj=lidj+myb0,bot=0,top=nnodes; 
	while(top-bot>5){
	  int i = (top+bot)/2;
	  if(gids[i] <= gidj) bot = i;
	  else top = i;
	}
	for(int i=bot;i<top;i++){
	  if( gids[i] == gidj ){ doit = 0; break; }
	  else if( gids[i] > gidj ) break;
	}
	if( doit == 0 ) { pa += bs2; continue; }
	px = &xarr[lidj*bs]; // local 
      	numops++;
	// b_i -= A_ij * x_j
	for( ii = 0 ; ii < bs ; ii++ ) {
	  tt = px[ii];
	  for( jj = 0 ; jj < bs ; jj++ ) tarr[jj] -= (*pa++) * tt;
	}
      }
      // inc ghosts (B)
      if( havgh && ghostx != NULL ) {
	pp = gh_ai[lidi]; pj = gh_aj + pp; 
	pa = gh_aa + bs2*pp; ncols = gh_ailen[lidi];
	// y = y - Ax (ghost)
	numops += ncols;
	for( colj = 0 ; colj != ncols ; colj++ ) {
	  int lidj = *pj++;
	  int gid = garr[lidj]; // PETSc's idea of global IDs
	  PromGSNode *node = gbid_ghost->find(gid+1);
	  if(node == NULL) { pa += bs2; continue; }
	  lidj = node->b.lid_;
	  px = &ghostx[lidj*bs]; 
	  // b_i -= A_ij * x_j
	  for( ii = 0 ; ii < bs ; ii++ ) {
	    tt = px[ii];
	    for( jj = 0 ; jj < bs ; jj++ ) tarr[jj] -= (*pa++) * tt;
	  }
	}
      }
      // y <- b_i - A_ij * x_j
      for(ii=0,jj=eq ; ii<bs ; ii++,jj++) yarr[jj] = tarr[ii];
    }
    ierr = VecRestoreArray( y, &yarr );       CHKERRQ(ierr);
    // apply -- do Aii^-1
    ierr = KSPSolve( ksp, y, x );    CHKERRQ(ierr); 
    // scatter ans
    ierr = VecGetArray( x, &pv );     CHKERRQ(ierr);  
    for( eq = 0 ; eq < nneq ; eq++ ) xarr[geqs[eq] - my0Eq] = pv[eq];
    ierr = VecRestoreArray( x, &pv );     CHKERRQ(ierr); 
    ierr = ISRestoreIndices( is, &geqs ); CHKERRQ(ierr);
  }
  ierr = xx->RestoreArray( &xarr );     CHKERRQ(ierr); 
  ierr = bb->RestoreArray( &barr );     CHKERRQ(ierr); 

  return 0;
}
   
#undef __FUNCT__  
#define __FUNCT__ "PromPCGaussSeidel::DoBlocks_schwarz_2"
int PromPCGaussSeidel::DoBlocks_schwarz_2( const int nb, const int blocks[],
					 const PromVector *const bb, 
					 PromVector *const xx,
					 const PromPC * const pc ) const
{
  if( nb == 0 ) return 0;
  int       ierr,ii,jj,colj,ncols,nnodes,eq,pp,nneq,numops,nn;
  const int *geqs;
  const int pre_s=pre_smooth_,*pj,kinc=(pre_s ? 1:-1),ksrt=(pre_s ? 0:nb-1);
  ierr = MatGetOwnershipRange( A_->mat_, &ii, &jj ); CHKERRQ( ierr );
  const int my0Eq=ii, myb0 = ii/bs_; 
  const bool havgh=(A_->map_.comm_.np() > 1);
  const double *aa=NULL,*gh_aa=NULL,*pa,*px;
  const int *ai,*aj,*ailen,*gh_ai=NULL,*gh_aj=NULL,*gh_ailen=NULL,*garr=NULL;
  Mat AA = A_->mat_; 
  Mat_SeqBAIJ *mat;
  int       gids[PROM_MAX_BLK_SZ];
  double   *barr,*yarr,*xarr,*pv,tarr[2];

  perf_mon_.EventBegin(PromContext::prom_logs_[LOG_FINALIZE]);

  if( havgh ) {
    Mat_MPIBAIJ *baij = (Mat_MPIBAIJ*)AA->data; // upper left matix
    garr = baij->garray;
    mat = (Mat_SeqBAIJ*)baij->B->data;
    // get ghost pointers
    gh_ai = mat->i; gh_aj = mat->j; gh_ailen = mat->ilen;   gh_aa = mat->a;
    //
    mat = (Mat_SeqBAIJ*)baij->A->data; 
  }
  else {
    mat = (Mat_SeqBAIJ*)AA->data;
  }
  // get local pointers
  ai = mat->i; aj = mat->j; ailen = mat->ilen;   aa = mat->a;

  ierr = bb->GetArray( &barr );       CHKERRQ(ierr);
  ierr = xx->GetArray( &xarr );       CHKERRQ(ierr);
  numops = 0;
  PETSc_ASM *osm = (PETSc_ASM*)pc_->pc_->data;             assert(osm != NULL);
  for( int xxx = 0, kk = ksrt ; xxx < nb ; xxx++, kk += kinc) {
    const int bidi = blocks[kk];
    // get rows of block
    IS is = osm->is[bidi];
    KSP ksp = osm->ksp[bidi];
    Vec x = osm->x[bidi]; Vec y = osm->y[bidi];
    ierr = VecGetArray( y, &yarr );       CHKERRQ(ierr);
    ierr = ISGetSize( is, &nneq ); CHKERRQ(ierr); nnodes = nneq/2;
    if(nnodes>PROM_MAX_BLK_SZ)SETERRQ2(1,"PROM_MAX_BLK_SZ(=%d) < bsz=%d",
				       PROM_MAX_BLK_SZ,nnodes);
    ierr = ISGetIndices( is, &geqs );                       CHKERRQ(ierr);
    // get gids (mask)
    for( nn = eq = 0 ; nn < nnodes ; nn++, eq += 2) gids[nn] = geqs[eq]/2;
    // y <- b - Ax
    for( nn = eq = 0 ; nn < nnodes ; nn++, eq += 2 ) {
      const int geq = geqs[eq], gidi = geq/2, lidi = gidi - myb0; 
      // get part (A) and 'bb' into temp array 'tarr'  
      if( lidi < 0 ) { // ghost
	SETERRQ2(1,"non processor nested domains not supported, lid=%d eq=%d",
		 lidi,geq);
      }
      else{
	pp = ai[lidi]; pj = aj + pp; // part (A)
	pa = aa + 4*pp;  ncols = ailen[lidi]; 
	px = &barr[lidi*2]; 
      }
      // tarr <- b
      for( ii = 0 ; ii < 2 ; ii++ ) tarr[ii] = px[ii]; 
      // y = y - Ax (local)
      for( colj = 0 ; colj != ncols ; colj++ ) {
	int lidj = *pj++;
	int doit=1,gidj=lidj+myb0,bot=0,top=nnodes; 
	while(top-bot>5){
	  int i = (top+bot)/2;
	  if(gids[i] <= gidj) bot = i;
	  else top = i;
	}
	for(int i=bot;i<top;i++){
	  if( gids[i] == gidj ){ doit = 0; break; }
	  else if( gids[i] > gidj ) break;
	}
	if( doit == 0 ) { pa += 4; continue; }
	px = &xarr[lidj*2]; // local 
	numops++;
	// b_i -= A_ij * x_j
	double tt,p0,p1;
	tt=px[0]; p0 = (*pa++)*tt; p1 = (*pa++)*tt; 
	tt=px[1]; p0 +=(*pa++)*tt; p1 +=(*pa++)*tt; 
	tarr[0] -= p0; tarr[1] -= p1;
      }
      if( havgh ) {
	pp = gh_ai[lidi]; pj = gh_aj + pp; 
	pa = gh_aa + 4*pp; ncols = gh_ailen[lidi];
	// y = y - Ax (ghost)
	for( colj = 0 ; colj != ncols ; colj++ ) {
	  int lidj = *pj++;
	  int gid = garr[lidj]; // PETSc's idea of global IDs
	  PromGSNode *node = gbid_ghost_.find(gid+1); 
	  if(node == NULL) { pa += 4; continue; }
	  numops++;
	  lidj = node->b.lid_;
	  px = &ghostx_[lidj*2]; 
	  // b_i -= A_ij * x_j
	  double tt,p0,p1;
	  tt=px[0]; p0 = (*pa++)*tt; p1 = (*pa++)*tt; 
	  tt=px[1]; p0 +=(*pa++)*tt; p1 +=(*pa++)*tt; 
	  tarr[0] -= p0; tarr[1] -= p1;
	}
      }
      // y <- b_i - A_ij * x_j
      for(ii=0,jj=eq ; ii<2 ; ii++,jj++) yarr[jj] = tarr[ii];
    }
    ierr = VecRestoreArray( y, &yarr );       CHKERRQ(ierr);
    // apply -- do Aii^-1
    ierr = KSPSolve( ksp, x, y );    CHKERRQ(ierr); 
    // scatter ans
    ierr = VecGetArray( x, &pv );     CHKERRQ(ierr);  
    for( eq = 0 ; eq < nneq ; eq++ ) xarr[geqs[eq] - my0Eq] = pv[eq];
    ierr = VecRestoreArray( x, &pv );     CHKERRQ(ierr); 
    ierr = ISRestoreIndices( is, &geqs ); CHKERRQ(ierr);
  }
  ierr = xx->RestoreArray( &xarr );     CHKERRQ(ierr); 
  ierr = bb->RestoreArray( &barr );     CHKERRQ(ierr); 
  //
  perf_mon_.Flops(2*4*numops);
  perf_mon_.EventEnd(PromContext::prom_logs_[LOG_FINALIZE]);
   
  return 0;
}
   
#undef __FUNCT__  
#define __FUNCT__ "PromPCGaussSeidel::DoBlocks_schwarz_3"
int PromPCGaussSeidel::DoBlocks_schwarz_3( const int nb, const int blocks[],
					 const PromVector *const bb, 
					 PromVector *const xx,
					 const PromPC * const pc ) const
{
  if( nb == 0 ) return 0;
  int       ierr,ii,jj,colj,ncols,nnodes,eq,pp,nneq,numops,nn;
  const int *geqs;
  const int pre_s=pre_smooth_,*pj,kinc=(pre_s ? 1:-1),ksrt=(pre_s ? 0:nb-1);
  ierr = MatGetOwnershipRange( A_->mat_, &ii, &jj ); CHKERRQ( ierr );
  const int my0Eq=ii, myb0 = ii/bs_; 
  const bool havgh=(A_->map_.comm_.np() > 1);
  const double *aa=NULL,*gh_aa=NULL,*pa,*px;
  const int *ai,*aj,*ailen,*gh_ai=NULL,*gh_aj=NULL,*gh_ailen=NULL,*garr=NULL;
  Mat AA = A_->mat_; 
  Mat_SeqBAIJ *mat;
  int       gids[PROM_MAX_BLK_SZ];
  double   *barr,*yarr,*xarr,*pv,tarr[3];

  perf_mon_.EventBegin(PromContext::prom_logs_[LOG_FINALIZE]);

  if( havgh ) {
    Mat_MPIBAIJ *baij = (Mat_MPIBAIJ*)AA->data; // upper left matix
    garr = baij->garray;
    mat = (Mat_SeqBAIJ*)baij->B->data;
    // get ghost pointers
    gh_ai = mat->i; gh_aj = mat->j; gh_ailen = mat->ilen;   gh_aa = mat->a;
    //
    mat = (Mat_SeqBAIJ*)baij->A->data; 
  }
  else {
    mat = (Mat_SeqBAIJ*)AA->data;
  }
  // get local pointers
  ai = mat->i; aj = mat->j; ailen = mat->ilen;   aa = mat->a;

  ierr = bb->GetArray( &barr );       CHKERRQ(ierr);
  ierr = xx->GetArray( &xarr );       CHKERRQ(ierr);
  numops = 0;
  PETSc_ASM *osm = (PETSc_ASM*)pc_->pc_->data;             assert(osm != NULL);
  for( int xxx = 0, kk = ksrt ; xxx < nb ; xxx++, kk += kinc) {
    const int bidi = blocks[kk];
    // get rows of block
    IS is = osm->is[bidi];
    KSP ksp = osm->ksp[bidi];
    Vec x = osm->x[bidi]; Vec y = osm->y[bidi];
    ierr = VecGetArray( y, &yarr );       CHKERRQ(ierr);
    ierr = ISGetSize( is, &nneq ); CHKERRQ(ierr); nnodes = nneq/3;
    if(nnodes>PROM_MAX_BLK_SZ)SETERRQ2(1,"PROM_MAX_BLK_SZ(=%d) < bsz=%d",
				       PROM_MAX_BLK_SZ,nnodes);
    ierr = ISGetIndices( is, &geqs );                       CHKERRQ(ierr);
    // get gids (mask)
    for( nn = eq = 0 ; nn < nnodes ; nn++, eq += 3) gids[nn] = geqs[eq]/3;
    // y <- b - Ax
    for( nn = eq = 0 ; nn < nnodes ; nn++, eq += 3 ) {
      const int geq = geqs[eq], gidi = geq/3, lidi = gidi - myb0; 
      // get part (A) and 'bb' into temp array 'tarr'  
      if( lidi < 0 ) { // ghost
	SETERRQ2(1,"non processor nested domains not supported, lid=%d eq=%d",
		 lidi,geq);
      }
      else{
	pp = ai[lidi]; pj = aj + pp; // part (A)
	pa = aa + 9*pp;  ncols = ailen[lidi]; 
	px = &barr[lidi*3]; 
      }
      // tarr <- b
      for( ii = 0 ; ii < 3 ; ii++ ) tarr[ii] = px[ii]; 
      // y = y - Ax (local)
      for( colj = 0 ; colj != ncols ; colj++ ) {
	int lidj = *pj++;
	int doit=1,gidj=lidj+myb0,bot=0,top=nnodes; 
	while(top-bot>5){
	  int i = (top+bot)/2;
	  if(gids[i] <= gidj) bot = i;
	  else top = i;
	}
	for(int i=bot;i<top;i++){
	  if( gids[i] == gidj ){ doit = 0; break; }
	  else if( gids[i] > gidj ) break;
	}
	if( doit == 0 ) { pa += 9; continue; }
	px = &xarr[lidj*3]; // local 
	numops++;
	// b_i -= A_ij * x_j
	double tt,p0,p1,p2;
	tt=px[0]; p0 = (*pa++)*tt; p1 = (*pa++)*tt; p2  = (*pa++)*tt;
	tt=px[1]; p0 +=(*pa++)*tt; p1 +=(*pa++)*tt; p2 += (*pa++)*tt;
	tt=px[2]; p0 +=(*pa++)*tt; p1 +=(*pa++)*tt; p2 += (*pa++)*tt;
	/*tt=px[0]; p0 = pa[0]*tt;  p1 = pa[1]*tt;  p2  = pa[2]*tt;
	  tt=px[1]; p0 += pa[3]*tt; p1 += pa[4]*tt; p2 += pa[5]*tt;
	  tt=px[2]; p0 += pa[6]*tt; p1 += pa[7]*tt; p2 += pa[8]*tt;
	  pa += 9;*/
	tarr[0] -= p0; tarr[1] -= p1;  tarr[2] -= p2; 
      }
      if( havgh ) {
	pp = gh_ai[lidi]; pj = gh_aj + pp; 
	pa = gh_aa + 9*pp; ncols = gh_ailen[lidi];
	// y = y - Ax (ghost)
	for( colj = 0 ; colj != ncols ; colj++ ) {
	  int lidj = *pj++;
	  int gid = garr[lidj]; // PETSc's idea of global IDs
	  PromGSNode *node = gbid_ghost_.find(gid+1); 
	  if(node == NULL) { pa += 9; continue; }
	  numops++;
	  lidj = node->b.lid_;
	  px = &ghostx_[lidj*3]; 
	  // b_i -= A_ij * x_j
	  double tt,p0,p1,p2;
	  tt=px[0]; p0 = (*pa++)*tt; p1 = (*pa++)*tt; p2  = (*pa++)*tt;
	  tt=px[1]; p0 +=(*pa++)*tt; p1 +=(*pa++)*tt; p2 += (*pa++)*tt;
	  tt=px[2]; p0 +=(*pa++)*tt; p1 +=(*pa++)*tt; p2 += (*pa++)*tt;
	  /*tt=px[0]; p0 = pa[0]*tt;  p1 = pa[1]*tt;  p2  = pa[2]*tt;
	    tt=px[1]; p0 += pa[3]*tt; p1 += pa[4]*tt; p2 += pa[5]*tt;
	    tt=px[2]; p0 += pa[6]*tt; p1 += pa[7]*tt; p2 += pa[8]*tt;
	    pa += 9;*/
	  tarr[0] -= p0; tarr[1] -= p1;  tarr[2] -= p2; 
	}
      }
      // y <- b_i - A_ij * x_j
      for(ii=0,jj=eq ; ii<3 ; ii++,jj++) yarr[jj] = tarr[ii];
    }
    ierr = VecRestoreArray( y, &yarr );       CHKERRQ(ierr);
    // apply -- do Aii^-1
    ierr = KSPSolve( ksp, y, x );    CHKERRQ(ierr); 
    // scatter ans
    ierr = VecGetArray( x, &pv );     CHKERRQ(ierr);  
    for( eq = 0 ; eq < nneq ; eq++ ) xarr[geqs[eq] - my0Eq] = pv[eq];
    ierr = VecRestoreArray( x, &pv );     CHKERRQ(ierr); 
    ierr = ISRestoreIndices( is, &geqs ); CHKERRQ(ierr);
  }
  ierr = xx->RestoreArray( &xarr );     CHKERRQ(ierr); 
  ierr = bb->RestoreArray( &barr );     CHKERRQ(ierr); 
  //
  perf_mon_.Flops(2*9*numops);
  perf_mon_.EventEnd(PromContext::prom_logs_[LOG_FINALIZE]);
   
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromPCGaussSeidel::DoBlocks_schwarz_6"
int PromPCGaussSeidel::DoBlocks_schwarz_6( const int nb, const int blocks[],
					   const PromVector *const bb, 
					   PromVector *const xx,
					   const PromPC * const pc ) const
{
  if( nb == 0 ) return 0;                        
  int       ierr,ii,jj,colj,ncols,nnodes,eq,pp,nneq,numops,nn;
  const int *geqs;
  const int pre_s=pre_smooth_,*pj,kinc=(pre_s ? 1:-1),ksrt=(pre_s ? 0:nb-1);
  ierr = MatGetOwnershipRange( A_->mat_, &ii, &jj ); CHKERRQ( ierr );
  const int my0Eq = ii, myb0 = ii/bs_;
  const bool havgh = (A_->map_.comm_.np() > 1);
  const double *aa=NULL,*gh_aa=NULL,*pa,*px;
  const int *ai,*aj,*ailen,*gh_ai=NULL,*gh_aj=NULL,*gh_ailen=NULL,*garr=NULL;
  Mat AA = A_->mat_; 
  Mat_SeqBAIJ *mat;
  int       gids[PROM_MAX_BLK_SZ];
  double   *barr,*xarr,*yarr,*pv,tarr[6];
  
  perf_mon_.EventBegin(PromContext::prom_logs_[LOG_FINALIZE]);
  if( havgh ) {
    Mat_MPIBAIJ *baij = (Mat_MPIBAIJ*)AA->data; // upper left matix
    garr = baij->garray;
    mat = (Mat_SeqBAIJ*)baij->B->data;
    // get ghost pointers
    gh_ai = mat->i; gh_aj = mat->j; gh_ailen = mat->ilen;   gh_aa = mat->a;
    //
    mat = (Mat_SeqBAIJ*)baij->A->data; 
  }
  else {
    mat = (Mat_SeqBAIJ*)AA->data;
  }
  // get local pointers
  ai = mat->i; aj = mat->j; ailen = mat->ilen;   aa = mat->a;
  
  ierr = bb->GetArray( &barr );       CHKERRQ(ierr);
  ierr = xx->GetArray( &xarr );       CHKERRQ(ierr);
  numops = 0;
  PETSc_ASM *osm = (PETSc_ASM*)pc_->pc_->data;             assert(osm != NULL);
  for( int xxx = 0, kk = ksrt ; xxx < nb ; xxx++, kk += kinc) {
    const int bidi = blocks[kk];
    // get rows of block
    IS is = osm->is[bidi];  KSP ksp = osm->ksp[bidi];
    Vec x = osm->x[bidi], y = osm->y[bidi];
    ierr = ISGetSize( is, &nneq ); CHKERRQ(ierr); nnodes = nneq/6;
    ierr = VecGetArray( y, &yarr );           //CHKERRQ(ierr);
    if(ierr)SETERRQ4(1,"Y errror %d/%d y=%p nneq=%d",xxx,nb,y,nneq);
    if(nnodes>PROM_MAX_BLK_SZ)SETERRQ2(1,"PROM_MAX_BLK_SZ(=%d) < bsz=%d",
				       PROM_MAX_BLK_SZ,nnodes);
    ierr = ISGetIndices( is, &geqs );                       CHKERRQ(ierr);
    // get gids (mask)
    for( nn = eq = 0 ; nn < nnodes ; nn++, eq += 6) gids[nn] = geqs[eq]/6;
    // y <- b - Ax
    for( nn = eq = 0 ; nn < nnodes ; nn++, eq += 6 ) {
      const int geq = geqs[eq], gidi = geq/6, lidi = gidi - myb0; 
      // get part (A) and 'bb' into temp array 'tarr'  
      if( lidi < 0 ) { // ghost
	SETERRQ1(1,"non processor nested domains not supported (lid=%d)",lidi);
      }
      else{
	pp = ai[lidi]; pj = aj + pp; // part (A)
	pa = aa + 36*pp;  ncols = ailen[lidi]; 
	px = &barr[lidi*6]; 
      }
      // tarr <- b
      for(ii=0 ; ii < 6 ; ii++) tarr[ii] = px[ii];
      // y = y - Ax (local)
      for( colj = 0 ; colj != ncols ; colj++ ) {
	int lidj = *pj++;
	int doit=1,gidj=lidj+myb0,bot=0,top=nnodes; 
	while(top-bot>5){
	  int i = (top+bot)/2;
	  if(gids[i] <= gidj) bot = i;
	  else top = i;
	}
	for(int i=bot;i<top;i++){
	  if( gids[i] == gidj ){ doit = 0; break; }
	  else if( gids[i] > gidj ) break;
	}
	if( doit == 0 ) { pa += 36; continue; }
	px = &xarr[lidj*6]; // local 
	numops++;
	// b_i -= A_ij * x_j
	double tt,p0,p1,p2,p3,p4,p5;
	tt = px[0]; 
	p0 = (*pa++)*tt; p1 = (*pa++)*tt; p2 = (*pa++)*tt;
	p3 = (*pa++)*tt; p4 = (*pa++)*tt; p5 = (*pa++)*tt;
	tt = px[1]; 
	p0 += (*pa++)*tt; p1 += (*pa++)*tt; p2 += (*pa++)*tt;
	p3 += (*pa++)*tt; p4 += (*pa++)*tt; p5 += (*pa++)*tt;
	tt = px[2]; 
	p0 += (*pa++)*tt;p1 += (*pa++)*tt;p2 += (*pa++)*tt;
	p3 += (*pa++)*tt;p4 += (*pa++)*tt;p5 += (*pa++)*tt;
	tt = px[3]; 
	p0 += (*pa++)*tt;p1 += (*pa++)*tt;p2 += (*pa++)*tt;
	p3 += (*pa++)*tt;p4 += (*pa++)*tt;p5 += (*pa++)*tt;
	tt = px[4]; 
	p0 += (*pa++)*tt;p1 += (*pa++)*tt;p2 += (*pa++)*tt;
	p3 += (*pa++)*tt;p4 += (*pa++)*tt;p5 += (*pa++)*tt;
	tt = px[5]; 
	p0 += (*pa++)*tt;p1 += (*pa++)*tt;p2 += (*pa++)*tt;
	p3 += (*pa++)*tt;p4 += (*pa++)*tt;p5 += (*pa++)*tt;
	tarr[0] -= p0; tarr[1] -= p1; tarr[2] -= p2; 
	tarr[3] -= p3; tarr[4] -= p4; tarr[5] -= p5;
      }
      if( havgh ) {
	pp = gh_ai[lidi]; pj = gh_aj + pp; 
	pa = gh_aa + 36*pp; ncols = gh_ailen[lidi];
	// y = y - Ax (ghost)
	numops += ncols;
	for( colj = 0 ; colj != ncols ; colj++ ) {
	  int lidj = *pj++;
	  int gid = garr[lidj]; // PETSc's idea of global IDs
	  PromGSNode *node = gbid_ghost_.find(gid+1); 
	  if(node == NULL) { pa += 36; continue; }
	  lidj = node->b.lid_;
	  px = &ghostx_[lidj*6]; 
	  // b_i -= A_ij * x_j
	  double tt,p0,p1,p2,p3,p4,p5;
	  tt = px[0]; 
	  p0 = (*pa++)*tt; p1 = (*pa++)*tt; p2 = (*pa++)*tt;
	  p3 = (*pa++)*tt; p4 = (*pa++)*tt; p5 = (*pa++)*tt;
	  tt = px[1]; 
	  p0 += (*pa++)*tt; p1 += (*pa++)*tt; p2 += (*pa++)*tt;
	  p3 += (*pa++)*tt; p4 += (*pa++)*tt; p5 += (*pa++)*tt;
	  tt = px[2]; 
	  p0 += (*pa++)*tt;p1 += (*pa++)*tt;p2 += (*pa++)*tt;
	  p3 += (*pa++)*tt;p4 += (*pa++)*tt;p5 += (*pa++)*tt;
	  tt = px[3]; 
	  p0 += (*pa++)*tt;p1 += (*pa++)*tt;p2 += (*pa++)*tt;
	  p3 += (*pa++)*tt;p4 += (*pa++)*tt;p5 += (*pa++)*tt;
	  tt = px[4]; 
	  p0 += (*pa++)*tt;p1 += (*pa++)*tt;p2 += (*pa++)*tt;
	  p3 += (*pa++)*tt;p4 += (*pa++)*tt;p5 += (*pa++)*tt;
	  tt = px[5]; 
	  p0 += (*pa++)*tt;p1 += (*pa++)*tt;p2 += (*pa++)*tt;
	  p3 += (*pa++)*tt;p4 += (*pa++)*tt;p5 += (*pa++)*tt;
	  tarr[0] -= p0; tarr[1] -= p1; tarr[2] -= p2; 
	  tarr[3] -= p3; tarr[4] -= p4; tarr[5] -= p5;
	}
      }
      // y <- b_i - A_ij * x_j
      for(ii=0,jj=eq ; ii<6 ; ii++,jj++) yarr[jj] = tarr[ii];
    }
    ierr = VecRestoreArray( y, &yarr );       CHKERRQ(ierr);
    // apply -- do Aii^-1
    ierr = KSPSolve( ksp, y, x );    CHKERRQ(ierr); 
    // scatter ans
    ierr = VecGetArray( x, &pv );     CHKERRQ(ierr);  
    for( eq = 0 ; eq < nneq ; eq++ ) xarr[geqs[eq] - my0Eq] = pv[eq];
    ierr = VecRestoreArray( x, &pv );     CHKERRQ(ierr); 
    ierr = ISRestoreIndices( is, &geqs ); CHKERRQ(ierr);
  }
  ierr = xx->RestoreArray( &xarr );     CHKERRQ(ierr); 
  ierr = bb->RestoreArray( &barr );     CHKERRQ(ierr); 
  //
  perf_mon_.Flops(2*36*numops);
  perf_mon_.EventEnd(PromContext::prom_logs_[LOG_FINALIZE]);
   
  return 0;
}
 
#undef __FUNCT__  
#define __FUNCT__ "PromPCGaussSeidel::DoBlocks_schwarz_1"
int PromPCGaussSeidel::DoBlocks_schwarz_1( const int nb, const int blocks[],
					   const PromVector *const bb, 
					   PromVector *const xx,
					   const PromPC * const pc ) const
{
  if( nb == 0 ) return 0;                        
  int       ierr,colj,ncols,nnodes,eq,pp,nneq,numops,nn;
  const int *geqs;
  const int pre_s=pre_smooth_,*pj,kinc=(pre_s ? 1:-1),ksrt=(pre_s ? 0:nb-1);
  ierr = MatGetOwnershipRange( A_->mat_, &eq, &nn ); CHKERRQ( ierr );
  const int my0Eq = eq, myb0 = eq/bs_;
  const bool havgh = (A_->map_.comm_.np() > 1);
  const double *aa=NULL,*gh_aa=NULL,*pa,*px;
  const int *ai,*aj,*ailen,*gh_ai=NULL,*gh_aj=NULL,*gh_ailen=NULL,*garr=NULL;
  Mat AA = A_->mat_; 
  Mat_SeqAIJ *mat;
  int       gids[PROM_MAX_BLK_SZ];
  double   *barr,*xarr,*yarr,*pv,tarr;
  
  perf_mon_.EventBegin(PromContext::prom_logs_[LOG_FINALIZE]);
  if( havgh ) {
    Mat_MPIAIJ *aij = (Mat_MPIAIJ*)AA->data; // upper left matix
    garr = aij->garray;
    mat = (Mat_SeqAIJ*)aij->B->data;
    // get ghost pointers
    gh_ai = mat->i; gh_aj = mat->j; gh_ailen = mat->ilen;   gh_aa = mat->a;
    //
    mat = (Mat_SeqAIJ*)aij->A->data; 
  }
  else {
    mat = (Mat_SeqAIJ*)AA->data;
  }
  // get local pointers
  ai = mat->i; aj = mat->j; ailen = mat->ilen;   aa = mat->a;
  ierr = bb->GetArray( &barr );       CHKERRQ(ierr);
  ierr = xx->GetArray( &xarr );       CHKERRQ(ierr);
  PETSc_ASM *osm = (PETSc_ASM*)pc_->pc_->data;             assert(osm != NULL);
  for( int xxx = numops = 0, kk = ksrt ; xxx < nb ; xxx++, kk += kinc) {
    const int bidi = blocks[kk];
    // get rows of block
    IS is = osm->is[bidi];
    KSP ksp = osm->ksp[bidi];
    Vec x = osm->x[bidi]; Vec y = osm->y[bidi];
    ierr = VecGetArray( y, &yarr );       CHKERRQ(ierr);
    ierr = ISGetSize( is, &nneq ); CHKERRQ(ierr); nnodes = nneq;
    if(nnodes>PROM_MAX_BLK_SZ)SETERRQ2(1,"PROM_MAX_BLK_SZ(=%d) < bsz=%d",
				       PROM_MAX_BLK_SZ,nnodes);
    ierr = ISGetIndices( is, &geqs );                       CHKERRQ(ierr);
    // get gids (mask)
    for( nn = eq = 0 ; nn < nnodes ; nn++, eq += 1 ) gids[nn] = geqs[eq];
    // y <- b - Ax
    for( nn = eq = 0 ; nn < nnodes ; nn++, eq += 1 ) {
      const int geq = geqs[eq], gidi = geq, lidi = gidi - myb0; 
      // get part (A) and 'bb' into temp array 'tarr'  
      if( lidi < 0 ) { // ghost
	SETERRQ1(1,"non processor nested domains not supported (lid=%d)",lidi);
      }
      else{
	pp = ai[lidi]; pj = aj + pp; // part (A)
	pa = aa + pp;  ncols = ailen[lidi]; 
	px = &barr[lidi]; 
      }
      // tarr <- b
      tarr = *px; 
      // y = y - Ax (local)
      for( colj = 0 ; colj != ncols ; colj++ ) {
	int lidj = *pj++;
	int doit=1,gidj=lidj+myb0,bot=0,top=nnodes; 
	while(top-bot>5){
	  int i = (top+bot)/2;
	  if(gids[i] <= gidj) bot = i;
	  else top = i;
	}
	for(int i=bot;i<top;i++){
	  if( gids[i] == gidj ){ doit = 0; break; }
	  else if( gids[i] > gidj ) break;
	}
	if( doit == 0 ) { pa += 1; continue; }
	px = &xarr[lidj]; // local 
	numops++;
	// b_i -= A_ij * x_j
	tarr -= (*pa++) * *px; 
      }
      if( havgh ) {
	pp = gh_ai[lidi]; pj = gh_aj + pp; 
	pa = gh_aa + pp; ncols = gh_ailen[lidi];
	// y = y - Ax (local,ghost)
	numops += ncols;
	for( colj = 0 ; colj != ncols ; colj++ ) {
	  int lidj = *pj++;
	  int gid = garr[lidj]; // PETSc's idea of global IDs
	  PromGSNode *node = gbid_ghost_.find(gid+1); 
	  if(node == NULL) { pa += 1; continue; }
	  lidj = node->b.lid_;
	  px = &ghostx_[lidj]; 
	  // b_i -= A_ij * x_j
	  tarr -= (*pa++) * *px; 
	}  
      }
      // y <- b_i - A_ij * x_j
      yarr[eq] = tarr;
    }
    ierr = VecRestoreArray( y, &yarr );       CHKERRQ(ierr);
    // apply -- do Aii^-1
    ierr = KSPSolve( ksp, y, x );    CHKERRQ(ierr); 
    // scatter ans
    ierr = VecGetArray( x, &pv );     CHKERRQ(ierr);  
    for( eq = 0 ; eq < nneq ; eq++ ) xarr[geqs[eq] - my0Eq] = pv[eq];
    ierr = VecRestoreArray( x, &pv );     CHKERRQ(ierr); 
    ierr = ISRestoreIndices( is, &geqs ); CHKERRQ(ierr);
  }
  ierr = xx->RestoreArray( &xarr );     CHKERRQ(ierr); 
  ierr = bb->RestoreArray( &barr );     CHKERRQ(ierr); 
  //
  perf_mon_.Flops(2*numops);
  perf_mon_.EventEnd(PromContext::prom_logs_[LOG_FINALIZE]);
   
  return 0;
}

#else

#undef __FUNCT__  
#define __FUNCT__ "PromPCGaussSeidel::DoBlocks-petra"
/* PromPCGaussSeidel::DoBlocks ***********************************
 *
 *   - apply blocks of GS smoothing w/o PETSc hacks -- nodal only 
 *
 *   INPUT:
 *    - nb
 *    - blocks[nb]: list of blocks (can be empty)
 *    - bb: 
 *    - xx; 
 *
 *   SIDE EFFECTS:
 * 
 *   RETURN:
 *     - PETSc error code
 */
int PromPCGaussSeidel::DoBlocks( const int nb, const int blocks[],
				 const PromVector * const bb,
				 PromVector *const xx ) const
{
  if( nb == 0 ) return 0;                       
  //if( block->arr_size_ == 0 ) return 0;
  int       ierr,ii,jj,colj,ncols;
  const int pre_s = pre_smooth_, fact = AA->isComplex() ? 2 : 1;
  ierr = MatGetOwnershipRange( A_->mat_, &ii, &jj ); CHKERRQ( ierr );
  const int my0eq = ii, nbloc = (jj-ii)/bs_;
  int       adjacs[PROM_MAXNIEGH],dirty,myb0=ii;
  int       numops,eq,neq,kinc=(pre_s ? 1 : -1);
  double    mAx[2*PROM_MAXNDF],*xarr,tt,p0,p1,p2,p3,p4,p5,*barr;
  double   *Ai[PROM_MAXNIEGH], *Aii;

  perf_mon_.EventBegin(LOG_FINALIZE);
  ierr = bb->GetArray( &barr );        CHKERRQ(ierr);
  ierr = xx->GetArray( &xarr );        CHKERRQ(ierr);
  for(int kkk = numops = 0, kk = (pre_s ? 0 : nb-1); kkk<nb; kkk++,kk += kinc){
    const int lidi = blocks[kk];  
    // nodal row
    int geqi = lidi*bs_, leqi = geqi - my0eq; 
    // y = y - Ax
    for( ii = 0 ; ii < bs_ ; ii++ ) mAx[ii] = 0.0;
    ierr = A_->GetLocal_constBlk_NodeRow(lidi, PROM_MAXNIEGH,ncols,adjacs,Ai);
    CHKERRQ(ierr);
    const int * pj = adjacs;   double * const *pa = Ai;
    for( colj = neq = 0, dirty = FALSE ; colj < ncols ; colj++ ) { 
      int jgid = *pj++, lidj = jgid - myb0;
      if( lidj == lidi ) continue; 
      const double *px;
      if( lidj < 0 || lidj >= nbloc ) { // ghost
	PromGSNode *node = gbid_ghost_.find(jgid+1); 
	if(node == NULL) continue;
	lidj = node->b.lid_;
	eq = lidj*bs_; 
	px = &ghostx_[eq]; 
      }
      else { // local 
	eq = lidj*bs_; 
	px = &xarr[eq]; 
      }
      neq += bs_;
      if( !dirty ) dirty = TRUE;
      // b_i -= A_ij * x_j
      const double *aijT = *pa++;
      switch( bs_ ){
      case 1: mAx[0] -= (*aijT++) * *px; break; 
      case 3:
	tt=px[0]; p0 = (*aijT++)*tt; p1 = (*aijT++)*tt; p2  = (*aijT++)*tt;
	tt=px[1]; p0 +=(*aijT++)*tt; p1 +=(*aijT++)*tt; p2 += (*aijT++)*tt;
	tt=px[2]; p0 +=(*aijT++)*tt; p1 +=(*aijT++)*tt; p2 += (*aijT++)*tt;
	mAx[0] -= p0; mAx[1] -= p1;  mAx[2] -= p2; 
	break; 
      case 6:
	tt = px[0]; 
	p0 = (*aijT++)*tt; p1 = (*aijT++)*tt; p2 = (*aijT++)*tt;
	p3 = (*aijT++)*tt; p4 = (*aijT++)*tt; p5 = (*aijT++)*tt;
	tt = px[1]; 
	p0 += (*aijT++)*tt; p1 += (*aijT++)*tt; p2 += (*aijT++)*tt;
	p3 += (*aijT++)*tt; p4 += (*aijT++)*tt; p5 += (*aijT++)*tt;
	tt = px[2]; 
	p0 += (*aijT++)*tt;p1 += (*aijT++)*tt;p2 += (*aijT++)*tt;
	p3 += (*aijT++)*tt;p4 += (*aijT++)*tt;p5 += (*aijT++)*tt;
	tt = px[3]; 
	p0 += (*aijT++)*tt;p1 += (*aijT++)*tt;p2 += (*aijT++)*tt;
	p3 += (*aijT++)*tt;p4 += (*aijT++)*tt;p5 += (*aijT++)*tt;
	tt = px[4]; 
	p0 += (*aijT++)*tt;p1 += (*aijT++)*tt;p2 += (*aijT++)*tt;
	p3 += (*aijT++)*tt;p4 += (*aijT++)*tt;p5 += (*aijT++)*tt;
	tt = px[5]; 
	p0 += (*aijT++)*tt;p1 += (*aijT++)*tt;p2 += (*aijT++)*tt;
	p3 += (*aijT++)*tt;p4 += (*aijT++)*tt;p5 += (*aijT++)*tt;
	mAx[0] -= p0; mAx[1] -= p1; mAx[2] -= p2; 
	mAx[3] -= p3; mAx[4] -= p4; mAx[5] -= p5;
	break; 
      default:
	for( ii = 0 ; ii < bs_ ; ii++ ) {
	  tt = px[ii];
	  for( jj = 0 ; jj < bs_ ; jj++ ) mAx[jj] -= (*aijT++) * tt;
	}
	break; 
      }
    }
    // b_i -= A_ij * x_j
    for(ii=0, jj = leqi ; ii < bs_ ; ii++, jj++ ) mAx[ii] += barr[jj];

    // do Aii^-1
    Aii = NodalFactors_ + NodalFactorPointers_[lidi];
    double tt,p0,p1,p2,p3,p4,p5;
    switch( bs_ ){
    case 3:
      tt = mAx[0]; p0 =  Aii[0]*tt; p1  = Aii[3]*tt; p2  = Aii[6]*tt;
      tt = mAx[1]; p0 += Aii[1]*tt; p1 += Aii[4]*tt; p2 += Aii[7]*tt;
      tt = mAx[2]; p0 += Aii[2]*tt; p1 += Aii[5]*tt; p2 += Aii[8]*tt;
      mAx[0] = p0; mAx[1] = p1; mAx[2] = p2; break; 
    case 6:
      tt = mAx[0]; 
      p0 = Aii[0]*tt;   p1 = Aii[6]*tt;   p2 = Aii[12]*tt;
      p3 = Aii[18]*tt;  p4 = Aii[24]*tt;  p5 = Aii[30]*tt;
      tt = mAx[1];
      p0 += Aii[1]*tt;  p1 += Aii[7]*tt;  p2 += Aii[13]*tt;
      p3 += Aii[19]*tt; p4 += Aii[25]*tt; p5 += Aii[31]*tt;
      tt = mAx[2]; 
      p0 += Aii[2]*tt;  p1 += Aii[8]*tt;  p2 += Aii[14]*tt;
      p3 += Aii[20]*tt; p4 += Aii[26]*tt; p5 += Aii[32]*tt;
      tt = mAx[3]; 
      p0 += Aii[3]*tt;  p1 += Aii[9]*tt;  p2 += Aii[15]*tt;
      p3 += Aii[21]*tt; p4 += Aii[27]*tt; p5 += Aii[33]*tt;
      tt = mAx[4]; 
      p0 += Aii[4]*tt;  p1 += Aii[10]*tt; p2 += Aii[16]*tt;
      p3 += Aii[22]*tt; p4 += Aii[28]*tt; p5 += Aii[34]*tt;
      tt = mAx[5]; 
      p0 += Aii[5]*tt;  p1 += Aii[11]*tt; p2 += Aii[17]*tt;
      p3 += Aii[23]*tt; p4 += Aii[29]*tt; p5 += Aii[35]*tt;
      mAx[0] = p0; mAx[1] = p1; mAx[2] = p2; 
      mAx[3] = p3; mAx[4] = p4; mAx[5] = p5;	break; 
    case 1: mAx[0] = Aii[0] * mAx[0];	break; 
    default :
      {
	double aa[100], *pv = Aii;
	for( ii = 0 ; ii < bs_ ; ii++ ) {
	  tt = 0.0; 
	  for( jj = 0 ; jj < bs_ ; jj++ ) tt += *pv++ *  mAx[jj];
	  aa[ii] = tt; 
	}
	for( ii = 0 ; ii < bs_ ; ii++ ) mAx[ii] = aa[ii];
      }
      break; 
    }
    for(ii=0, jj = leqi ; ii < bs_ ; ii++, jj++ ) xarr[jj] = mAx[ii];
    //ierr = xx->SetValues( fndf, tids, mAx, INSERT_VALUES ); CHKERRQ(ierr);
    numops += ncols * bs_ * bs_;
  } // kk subdomain
  ierr = xx->RestoreArray( &xarr );     CHKERRQ(ierr);
  ierr = bb->RestoreArray( &barr );     CHKERRQ(ierr);
  perf_mon_.Flops(2*numops);
  perf_mon_.EventEnd(PromContext::prom_logs_[LOG_FINALIZE]);

  return 0;
}

#endif
